import base64
import gzip
import html
import io
import os
import warnings
import zipfile
import json

import jinja2
import numpy as np
import pandas as pd
import requests
import re
from importlib_resources import files
from matplotlib.colors import to_rgba
from pathlib import Path
from rcssmin import cssmin
from rjsmin import jsmin
from scipy.spatial import Delaunay

from pandas.api.types import is_string_dtype, is_numeric_dtype, is_datetime64_any_dtype

from datamapplot.histograms import generate_bins_from_numeric_data, generate_bins_from_categorical_data, generate_bins_from_temporal_data
from datamapplot.alpha_shapes import create_boundary_polygons, smooth_polygon
from datamapplot.medoids import medoid

_DECKGL_TEMPLATE_STR = (files("datamapplot") / "deckgl_template.html").read_text(
    encoding="utf-8"
)

_TOOL_TIP_CSS = """
            font-size: 0.8em;
            font-family: {{title_font_family}};
            font-weight: {{title_font_weight}};
            color: {{title_font_color}} !important;
            background-color: {{title_background}} !important;
            border-radius: 12px;
            box-shadow: 2px 3px 10px {{shadow_color}};
            max-width: 25%;
"""

_NOTEBOOK_NON_INLINE_WORKER = """
    const parsingWorkerBlob = new Blob([`
      async function DecompressBytes(bytes) {
          const blob = new Blob([bytes]);
          const decompressedStream = blob.stream().pipeThrough(
            new DecompressionStream("gzip")
          );
          const arr = await new Response(decompressedStream).arrayBuffer()
          return new Uint8Array(arr);
      }
      async function decodeBase64(base64) {
          return Uint8Array.from(atob(base64), c => c.charCodeAt(0));
      }
      async function decompressFile(filename) {
          const response = await fetch(filename, {
            headers: {Authorization: 'Token API_TOKEN'}
          });
          if (!response.ok) {
            throw new Error(\`HTTP error! status: \${response.status}\`);
          }
          const data = await response.json()
            .then(data => data.content)
            .then(base64data => decodeBase64(base64data))
            .then(buffer => DecompressBytes(buffer));
          return data;
      }
      self.onmessage = async function(event) {
        const { encodedData, JSONParse } = event.data;
        const binaryData = await decompressFile(encodedData);
        if (JSONParse) {
          const parsedData = JSON.parse(new TextDecoder("utf-8").decode(binaryData));
          self.postMessage({ data: parsedData });
        } else {
          // Send the parsed table back to the main thread
          self.postMessage({ data: binaryData });
        }
      }
    `], { type: 'application/javascript' });
"""


class FormattingDict(dict):
    def __missing__(self, key):
        return f"{{{key}}}"


class InteractiveFigure:
    """A simple class to hold interactive plot outputs. This allows for
    and object with a `_repr_html_` to display in notebooks.

    The actual content of the plot is HTML, and the `save` method can
    be used to save the results to an HTML file while can then be shared.
    """

    def __init__(self, html_str, width="100%", height=800, api_token=None):
        self._html_str = html_str
        self.width = width
        self.height = height
        self.api_token = api_token or os.environ.get("JUPYTERHUB_API_TOKEN", None)

    def __repr__(self):
        return f"<InteractiveFigure width={self.width} height={self.height}>"

    def __str__(self):
        return self._html_str

    def _repr_html_(self):
        if "originURL" in self._html_str:
            # We need to redirect the fetch to use the jupyter API endpoint
            # for use in a notebook...
            jupyter_html_str = self._html_str.replace(
                "originURL = self.location.origin + directoryPath;", 
                "originURL = document.baseURI.substring(0, document.baseURI.lastIndexOf('/')).replace(/(notebooks|lab.*tree)/, 'api/contents');"
            )
            jupyter_html_str = re.sub(
                r'const parsingWorkerBlob.*?\'application/javascript\' \}\);', 
                _NOTEBOOK_NON_INLINE_WORKER,
                jupyter_html_str,
                flags=re.DOTALL
            )
            if self.api_token is not None:
                jupyter_html_str = jupyter_html_str.replace(
                    "headers: {Authorization: 'Token API_TOKEN'}",
                    f"headers: {{Authorization: 'Token {self.api_token}'}}"
                )
            else:
                jupyter_html_str = jupyter_html_str.replace(
                    "headers: {Authorization    : 'Token API_TOKEN'}",
                    ""
                )
            src_doc =  html.escape(jupyter_html_str)
        else:
            src_doc = html.escape(self._html_str)
        iframe = f"""
            <iframe
                width={self.width}
                height={self.height}
                frameborder="0"
                srcdoc="{src_doc}"
            ></iframe>
        """
        from IPython.display import HTML

        with warnings.catch_warnings():
            msg = "Consider using IPython.display.IFrame instead"
            warnings.filterwarnings("ignore", message=msg)

            html_obj = HTML(iframe)
            return getattr(html_obj, "data", "")

    def save(self, filename):
        """Save an interactive firgure to the HTML file with name `filename`"""
        with open(filename, "w+", encoding="utf-8") as f:
            f.write(self._html_str)

    def save_bundle(self, filename):
        """Save an interactive figure to a zip file with name `filename`"""
        with zipfile.ZipFile(filename, "w") as zf:
            zf.writestr("index.html", self._html_str)
            for filename in re.findall(r'/(.*?_data\.zip)', self._html_str):
                print(f"Adding {filename} to bundle")
                zf.write(filename)



def get_google_font_for_embedding(fontname):
    api_fontname = fontname.replace(" ", "+")
    api_response = requests.get(
        f"https://fonts.googleapis.com/css?family={api_fontname}:black,bold,regular,light",
        timeout=10,
    )
    if api_response.ok:
        font_urls = re.findall(r"(https?://[^\)]+)", str(api_response.content))
        font_links = []
        for url in font_urls:
            if url.endswith(".ttf"):
                font_links.append(
                    f'<link rel="preload" href="{url}" as="font" crossorigin="anonymous" type="font/ttf" />'
                )
            elif url.endswith(".woff2"):
                font_links.append(
                    f'<link rel="preload" href="{url}" as="font" crossorigin="anonymous" type="font/woff2" />'
                )
        return "\n".join(font_links) + f"\n<style>\n{api_response.content.decode()}\n</style>\n"
    else:
        return ""


def _get_js_dependency_sources(minify, enable_search, enable_histogram, enable_lasso_selection):
    """
    Gather the necessary JavaScript dependency files for embedding in the HTML template.

    Parameters
    ----------
    minify : bool
        Whether to minify the JS files.

    enable_search : bool
        Whether to include JS dependencies for the search functionality.

    enable_histogram: bool
        Whether to include JS dependencies for the histogram functionality.

    enable_lasso_selection: bool
        Whether to include JS dependencies for the lasso selection functionality.

    Returns
    -------
    dict
        A dictionary where keys are the names of JS files and values are their
        source content.
    """
    static_dir = Path(__file__).resolve().parent / "static" / "js"
    js_dependencies = ["datamap.js", "data_selection_manager.js"]
    js_dependencies_src = {}

    if enable_histogram:
        js_dependencies.append("d3_histogram.js")

    if enable_lasso_selection:
        js_dependencies.append("lasso_selection.js")
        js_dependencies.append("quad_tree.js")

    for js_file in js_dependencies:
        with open(static_dir / js_file, "r", encoding="utf-8") as file:
            js_src = file.read()
            js_dependencies_src[js_file] = jsmin(js_src) if minify else js_src

    return js_dependencies_src


def _get_css_dependency_sources(minify, enable_histogram, show_loading_progress):
    """
    Gather the necessary CSS dependency files for embedding in the HTML template.

    Parameters
    ----------
    minify : bool
        Whether to minify the CSS files.

    enable_histogram: bool
        Whether to include CSS dependencies for the histogram functionality.

    show_loading_progress: bool
        Whether to have progress bars for data loading.

    Returns
    -------
    dict
        A dictionary where keys are the names of CSS files and values are their
        source content.
    """
    static_dir = Path(__file__).resolve().parent / "static" / "css"
    css_dependencies = []
    css_dependencies_src = {}

    if enable_histogram:
        css_dependencies.append("d3_histogram_style.css")

    if show_loading_progress:
        css_dependencies.append("progress_bar_style.css")

    for css_file in css_dependencies:
        with open(static_dir / css_file, "r", encoding="utf-8") as file:
            css_src = file.read()
            css_dependencies_src[css_file] = cssmin(css_src) if minify else css_src

    return css_dependencies_src


def _get_js_dependency_urls(enable_histogram, selection_handler=None):
    """
    Gather the necessary JavaScript dependency URLs for embedding in the HTML template.

    Parameters
    ----------
    enable_histogram: bool
        Whether to include JS URLs for the histogram functionality.

    Returns
    -------
    list
        A list of URLs that point to the required JavaScript dependencies.
    """
    js_dependency_urls = []

    # Add common dependencies (if any)
    common_js_urls = [
        "https://unpkg.com/deck.gl@latest/dist.min.js",
        "https://unpkg.com/apache-arrow@latest/Arrow.es2015.min.js",
    ]
    js_dependency_urls.extend(common_js_urls)

    # Conditionally add dependencies based on functionality
    if enable_histogram:
        js_dependency_urls.append(
            "https://d3js.org/d3.v6.min.js"
        )

    if selection_handler is not None:
        js_dependency_urls.extend(selection_handler.dependencies)

    js_dependency_urls = list(set(js_dependency_urls))

    return js_dependency_urls


def label_text_and_polygon_dataframes(
    labels,
    data_map_coords,
    noise_label="Unlabelled",
    use_medoids=False,
    cluster_polygons=False,
    alpha=0.05,
):
    cluster_label_vector = np.asarray(labels)
    unique_non_noise_labels = [
        label for label in np.unique(cluster_label_vector) if label != noise_label
    ]
    label_map = {n: i for i, n in enumerate(unique_non_noise_labels)}
    label_map[noise_label] = -1
    cluster_idx_vector = np.asarray(pd.Series(cluster_label_vector).map(label_map))

    label_locations = []
    cluster_sizes = []
    polygons = []

    for i, _ in enumerate(unique_non_noise_labels):
        cluster_mask = cluster_idx_vector == i
        cluster_points = data_map_coords[cluster_mask]
        if use_medoids:
            label_locations.append(medoid(cluster_points))
        else:
            label_locations.append(cluster_points.mean(axis=0))

        cluster_sizes.append(np.sum(cluster_mask) ** 0.25)
        if cluster_polygons:
            simplices = Delaunay(cluster_points, qhull_options="Qbb Qc Qz Q12 Q7").simplices
            polygons.append(
                [
                    smooth_polygon(x).tolist()
                    for x in create_boundary_polygons(
                        cluster_points, simplices, alpha=alpha
                    )
                ]
            )

    label_locations = np.asarray(label_locations)

    data = {
        "x": label_locations.T[0],
        "y": label_locations.T[1],
        "label": unique_non_noise_labels,
        "size": cluster_sizes,
    }
    if cluster_polygons:
        data["polygon"] = polygons

    return pd.DataFrame(data)
    

def render_html(
    point_dataframe,
    label_dataframe,
    inline_data=True,
    title=None,
    sub_title=None,
    title_font_size=36,
    sub_title_font_size=18,
    text_collision_size_scale=3,
    text_min_pixel_size=18,
    text_max_pixel_size=36,
    font_family="Roboto",
    font_weight=600,
    tooltip_font_family=None,
    tooltip_font_weight=300,
    logo=None,
    logo_width=256,
    color_label_text=True,
    line_spacing=0.95,
    min_fontsize=18,
    max_fontsize=28,
    text_outline_width=8,
    text_outline_color="#eeeeeedd",
    point_size_scale=None,
    point_hover_color="#aa0000bb",
    point_radius_min_pixels=0.01,
    point_radius_max_pixels=24,
    point_line_width_min_pixels=0.001,
    point_line_width_max_pixels=3,
    point_line_width=0.001,
    cluster_boundary_line_width=1,
    initial_zoom_fraction=1.0,
    background_color=None,
    darkmode=False,
    offline_data_prefix=None,
    tooltip_css=None,
    hover_text_html_template=None,
    extra_point_data=None,
    enable_search=False,
    search_field="hover_text",
    histogram_data=None,
    histogram_n_bins=20,
    histogram_group_datetime_by=None,
    histogram_range=None,
    histogram_settings={},
    on_click=None,
    selection_handler=None,
    show_loading_progress=True,
    custom_html=None,
    custom_css=None,
    custom_js=None,
    minify_deps=True,
):
    """Given data about points, and data about labels, render to an HTML file
    using Deck.GL to provide an interactive plot that can be zoomed, panned
    and explored.

    Parameters
    ----------
    point_dataframe: pandas.DataFrame
        A Dataframe containing point information for rendering. At a minimum this
        should include columns `"x"`, `"y"`, `"r"`, `"g"`, `"b"` and `"a"` that
        provide the x,y position and r,g,b color for each point. Note that r,g,b,a
        values should be uint8 values.

    label_dataframe: pandas.DataFrame
        A Dataframe containing information about labels, and optionally bounding
        polygons, of clusters. At a minimum this should include columns
        `"x"`, `"y"`, `"r"`, `"g"`, `"b"` and `"a"` that provide the x,y position
        and r,g,b colour of the label.

    inline_data: bool (optional, default=True)
        Whether to include data inline in the HTML file (compressed and base64 encoded)
        of whether to write data to separate files that will then be referenced by the
        HTML file -- in the latter case you will need to ensure all the files are
        co-located and served over an http server or similar. Inline is the best
        default choice for easy portability and simplicity, but can result in very
        large file sizes.

    title: str or None (optional, default=None)
        A title for the plot, to be placed in the top left corner. The title should be
        brief and to the point. More detail can be provided in the sub_title if required.

    sub_title: str or None (optional, default=None)
        A sub_title for the plot, to be placed in the top left corner.

    title_font_size: int (optional, default=36)
        The font-size of the title in points.

    sub_title_font_size: int (optional, default=18)
        The font-size of the sub-title in points.

    text_collision_size_scale: float (optional, default=3.0)
        How to scale text labels for the purpose of collision detection to determine
        which labels to display.

    text_min_pixel_size: float (optional, default=12.0)
        The minimum pixel size of label text. If text would be smaller than this in size
        then render the text to be at least this size.

    text_max_pixel_size: float (optional, default=36.0)
        The maximum pixel size of label text. If text would be larger than this in size
        then render the text to be at most this size.

    font_family: str (optional, default="Roboto")
        The font family to use for label text and titles. If the font family is a
        google font then the required google font api handling will automatically
        make the font available, so any google font family is acceptable.

    font_weight: str or int (optional, default=600)
        The font weight to use for the text labels within the plot. Either weight
        specification such as "thin", "normal", or "bold" or an integer value
        between 0 (ultra-thin) and 1000 (ultra-black).

    tooltip_font_family: str (optional default="Roboto")
        The font family to use in tooltips/hover text. If the font family is a
        google font then the required google font api handling will automatically
        make the font available, so any google font family is acceptable.

    tooltip_font_weight: str or int (optional, default=400)
        The font weight to use for the tooltip /hover text within the plot. Either weight
        specification such as "thin", "normal", or "bold" or an integer value
        between 0 (ultra-thin) and 1000 (ultra-black).

    logo: str or None (optional, default=None)
        A logo image to include in the bottom right corner of the map. This should be
        a URL to the image.

    logo_width: int (optional, default=256)
        The width, in pixels, of the logo to be included in the bottom right corner.
        The logo will retain it's aspect ratio, so choose the width accordingly.

    color_label_text: bool (optional, default=True)
        Whether the text labels for clusters should be coloured or not. If set to False
        the labels will be either black or white depending on whether ``darkmode`` is set.

    line_spacing: float (optional, default=0.95)
        Line height spacing in label text.

    min_fontsize: float (optional, default=12)
        The minimum font size (in points) of label text. In general label text is scaled
        based on the size of the cluster the label if for; this will set the minimum
        value for that scaling.

    max_fontsize: float (optional, default=24)
        The maximum font size (in points) of label text. In general label text is scaled
        based on the size of the cluster the label if for; this will set the maximum
        value for that scaling.

    text_outline_width: float (optional, default=8)
        The size of the outline around the label text. The outline, in a contrasting
        colour, can make text more readable against the map background. Choosing larger
        sizes can help if text is less legible.

    text_outline_color: str (optional, default="#eeeeeedd")
        The colour of the outline around the label text. The outline should be a
        contrasting colour to the colour of the label text. By default this is white
        when ``darkmode`` is ``False`` and black when ``darkmode`` is ``True``.

    point_size_scale: float or None (optional, default=None)
        The size scale of points. If None the size scale will be determined from the data.

    point_hover_color: str (optional, default="#aa0000bb")
        The colour of the highlighted point a user is hovering over.

    point_radius_min_pixels: float (optional, default=0.01)
        The minimum number of pixels in radius of the points in the map; if zoomed out
        enough that a point would be smaller than this, it is instead rendered at this radius.
        This allows points to remain visible when zoomed out.

    point_radius_max_pixels: float (optional, default=24)
        The maximum number of pixels in radius of the points in the map; if zoomed in
        enough that a point would be larger than this, it is instead rendered at this radius.
        This allows zooming in to differentiate points that are otherwise overtop of one
         another.

    point_line_width_min_pixels: float (optional, default=0.001)
        The minimum pixel width of the outline around points.

    point_line_width_max_pixels: float (optional, default=3)
        The maximum pixel width of the outline around points.

    point_line_width: float (optional, default=0.001)
        The absolute line-width in common coordinates of the outline around points.

    cluster_boundary_line_width: float (optional, default=1.0)
        The linewidth to use for cluster boundaries. Note that cluster boundaries scale with respect
        to cluster size, so this is a scaling factor applied over this.

    initial_zoom_fraction: float (optional, default=1.0)
        The fraction of the total zoom (containing allm the data) to start the
        map in. A lower value will initialize the plot zoomed in, while values
        larger than 1.0 will result in the initial start point being zoomed out.

    background_color: str or None (optional, default=None)
        A background colour (as a hex-string) for the data map. If ``None`` a background
        colour will be chosen automatically based on whether ``darkmode`` is set.

    darkmode: bool (optional, default=False)
        Whether to use darkmode.

    offline_data_prefix: str or None (optional, default=None)
        If ``inline_data=False`` a number of data files will be created storing data for
        the plot and referenced by the HTML file produced. If not none then this will provide
        a prefix on the filename of all the files created.

    tooltip_css: str or None (optional, default=None)
        Custom CSS used to fine the properties of the tooltip. If ``None`` a default
        CSS style will be used. This should simply be the required CSS directives
        specific to the tooltip.

    hover_text_html_template: str or None (optional, default=None)
        An html template allowing fine grained control of what is displayed in the
        hover tooltip. This should be HTML with placeholders of the form ``{hover_text}``
        for the supplied hover text and ``{column_name}`` for columns from
        ``extra_point_data`` (see below).

    extra_point_data: pandas.DataFrame or None (optional, default=None)
        A dataframe of extra information about points. This should be a dataframe with
        one row per point. The information in this dataframe can be referenced by column-name
        by either ``hover_text_html_template`` or ``on_click`` for use in tooltips
        or on-click actions.

    enable_search: bool (optional, default=False)
        Whether to enable a text search that can highlight points with hover_text that
        include the given search string.

    search_field: str (optional, default="hover_text")
        If ``enable_search`` is ``True`` and ``extra_point_data`` is not ``None``, then search
        this column of the ``extra_point_data`` dataframe, or use hover_text if set to
        ``"hover_text"``.

    histogram_data: list, pandas.Series, or None (optional, default=None)
        The data used to generate a histogram. The histogram data can be passed as a list or
        Pandas Series; if `None`, the histogram is disabled. The length of the list or Series
        must match the number of rows in `point_dataframe`. The values within the list or Series
        must be of type unsigned integer, signed integer, floating-point number, string, or a
        date string in the format `YYYY-MM-DD`.

    histogram_n_bins: int (optional, default=20)
        The number of bins in the histogram. It is the maximum number of bins if binning categorical
        data. If the number of unique values in the data is less than or equal to `histogram_n_bins`,
        the number of bins will be the number of unique values.

    histogram_group_datetime_by: str or None (optional, default=None)
        The time unit to group the datetime data by. If `None`, the datetime data will not be
        grouped. The time unit can be one of the following: `year`, `quarter`, `month`, `week`,
        `day`, `hour`, `minute`, or `second`.

    histogram_range: tuple or None (optional, default=None)
        The range of the histogram. If `None`, the range is automatically determined from the   
        histogram data. If a tuple, it should contain two values representing the minimum and
        maximum values of the histogram.

    histogram_settings: dict or None (optional, default={})
        A dictionary containing custom settings for the histogram, if enabled. If
        `histogram_data` is provided, this dictionary allows you to customize the
        appearance of the histogram. The dictionary can include the following keys:

        - "histogram_width": str
            The width of the histogram in pixels.
        - "histogram_height": str
            The height of the histogram in pixels.
        - "histogram_bin_count": int
            The number of bins in the histogram.
        - "histogram_title": str
            The title of the histogram.
        - "histogram_bin_fill_color": str
            The fill HEX color of the histogram bins (e.g. `#6290C3`).
        - "histogram_bin_selected_fill_color": str
            The fill HEX color of the selected histogram bins (e.g. `#2EBFA5`).
        - "histogram_bin_unselected_fill_color": str
            The fill HEX color of the unselected histogram bins (e.g. `#9E9E9E`).
        - "histogram_bin_context_fill_color": str
            The fill HEX color of the contextual bins in the histogram (e.g. `#E6E6E6`).
        - "histogram_log_scale": bool
            Whether to use a log scale for y-axis of the histogram.

    on_click: str or None (optional, default=None)
        A javascript action to be taken if a point in the data map is clicked. The javascript
        can reference ``{hover_text}`` or columns from ``extra_point_data``. For example one
        could provide ``"window.open(`http://google.com/search?q=\"{hover_text}\"`)"`` to
        open a new window with a google search for the hover_text of the clicked point.

    selection_handler: instance of datamapplot.selection_handlers.SelectionHandlerBase or None (optional, default=None)
        A selection handler to be used to handle selections in the data map. If None, the
        interactive selection will not be enabled. If a selection handler is provided, the
        selection handler will be used to determine how to react to selections made on the
        data map. Selection handlers can be found in the `datamapplot.selection_handlers`
        module, or custom selection handlers can be created by subclassing the `SelectionHandlerBase`
        class.

    custom_css: str or None (optional, default=None)
        A string of custom CSS code to be added to the style header of the output HTML. This
        can be used to provide custom styling of other features of the output HTML as required.

    custom_html: str or None (optional, default=None)
        A string of custom HTML to be added to the body of the output HTML. This can be used to
        add other custom elements to the interactive plot, including elements that can be
        interacted with via the ``on_click`` action for example.

    custom_js: str or None (optional, default=None)
        A string of custom Javascript code that is to be added after the code for rendering
        the scatterplot. This can include code to interact with the plot which is stored
        as ``deckgl``.

    minify_deps: bool (optional, default=True)
        Whether to minify the JavaScript and CSS dependency files before embedding in the HTML template.

    Returns
    -------
    interactive_plot: InteractiveFigure
        An interactive figure with hover, pan, and zoom. This will display natively
        in a notebook, and can be saved to an HTML file via the `save` method.
    """
    # Compute point scaling
    n_points = point_dataframe.shape[0]
    if point_size_scale is not None:
        magic_number = point_size_scale / 100.0
    else:
        magic_number = np.clip(32 * 4 ** (-np.log10(n_points)), 0.005, 0.1)

    if "size" not in point_dataframe.columns:
        point_size = magic_number
    else:
        point_dataframe["size"] = magic_number * (
            point_dataframe["size"] / point_dataframe["size"].mean()
        )
        point_size = -1

    # Compute zoom level and initial view location
    data_width = point_dataframe.x.max() - point_dataframe.x.min()
    data_height = point_dataframe.y.max() - point_dataframe.y.min()
    data_center = point_dataframe[["x", "y"]].values.mean(axis=0)

    spread = max(data_width, data_height) * initial_zoom_fraction
    if spread < (360.0 * np.power(2.0, -20)):
        zoom_level = 21
    else:
        zoom_level = max(1, np.log2(360.0) - np.log2(spread))

    if darkmode and text_outline_color == "#eeeeeedd":
        text_outline_color = "#111111dd"

    point_outline_color = [250, 250, 250, 128] if not darkmode else [5, 5, 5, 128]
    text_background_color = [255, 255, 255, 64] if not darkmode else [0, 0, 0, 64]
    if color_label_text:
        label_text_color = "d => [d.r, d.g, d.b]"
    else:
        label_text_color = [0, 0, 0, 255] if not darkmode else [255, 255, 255, 255]

    # Compute text scaling
    size_range = label_dataframe["size"].max() - label_dataframe["size"].min()
    if size_range > 0:
        label_dataframe["size"] = (
            label_dataframe["size"] - label_dataframe["size"].min()
        ) * ((max_fontsize - min_fontsize) / size_range) + min_fontsize
    else:
        label_dataframe["size"] = (max_fontsize + min_fontsize) / 2.0

    # Prep data for inlining or storage
    enable_histogram = histogram_data is not None
    histogram_data_attr = "histogram_data_attr"
    histogram_ctx = {
        "enable_histogram": enable_histogram,
        "histogram_data_attr": histogram_data_attr,
        **histogram_settings,
    }
    enable_lasso_selection = selection_handler is not None        

    point_data_cols = ["x", "y", "r", "g", "b", "a"]

    if point_size < 0:
        point_data_cols.append("size")

    # if enable_search or enable_histogram or enable_lasso_selection:
    #     point_dataframe["selected"] = np.ones(len(point_dataframe), dtype=np.uint8)
    #     point_data_cols.append("selected")

    point_data = point_dataframe[point_data_cols]

    if "hover_text" in point_dataframe.columns:
        if extra_point_data is not None:
            hover_data = pd.concat(
                [point_dataframe[["hover_text"]], extra_point_data],
                axis=1,
            )
            replacements = FormattingDict(
                **{
                    str(name): f"${{hoverData.{name}[index]}}"
                    for name in hover_data.columns
                }
            )
            if hover_text_html_template is not None:
                get_tooltip = (
                    '({index, picked}) => picked ? {"html": `'
                    + hover_text_html_template.format_map(replacements)
                    + "`} : null"
                )
            else:
                get_tooltip = "({index}) => hoverData.hover_text[index]"

            if on_click is not None:
                on_click = (
                    "({index, picked}, event) => { if (picked) {"
                    + on_click.format_map(replacements)
                    + " } }"
                )
        else:
            hover_data = point_dataframe[["hover_text"]].copy()
            get_tooltip = "({index}) => hoverData.hover_text[index]"

            replacements = FormattingDict(
                **{
                    str(name): f"${{hoverData.{name}[index]}}"
                    for name in hover_data.columns
                }
            )

            if on_click is not None:
                on_click = (
                    "({index, picked}, event) => { if (picked) {"
                    + on_click.format_map(replacements)
                    + " } }"
                )
    elif extra_point_data is not None:
        hover_data = extra_point_data.copy()
        replacements = FormattingDict(
            **{
                str(name): f"${{hoverData.{name}[index]}}"
                for name in hover_data.columns
            }
        )
        if hover_text_html_template is not None:
            get_tooltip = (
                '({index, picked}) => picked ? {"html": `'
                + hover_text_html_template.format_map(replacements)
                + "`} : null"
            )
        else:
            get_tooltip = "null"

        if on_click is not None:
            on_click = (
                "({index, picked}, event) => { if (picked) {"
                + on_click.format_map(replacements)
                + " } }"
            )
    else:
        hover_data = pd.DataFrame(columns=("hover_text",))
        get_tooltip = "null"

    if enable_histogram:
        if isinstance(histogram_data.dtype, pd.CategoricalDtype):
            bin_data, index_data = generate_bins_from_categorical_data(histogram_data, histogram_n_bins, histogram_range)
        elif is_string_dtype(histogram_data.dtype):
            bin_data, index_data = generate_bins_from_categorical_data(histogram_data, histogram_n_bins, histogram_range)
        elif is_datetime64_any_dtype(histogram_data.dtype):
            if histogram_group_datetime_by is not None:
                bin_data, index_data = generate_bins_from_temporal_data(histogram_data, histogram_group_datetime_by, histogram_range)
            else:
                bin_data, index_data = generate_bins_from_numeric_data(histogram_data, histogram_n_bins, histogram_range)
        else:
            bin_data, index_data = generate_bins_from_numeric_data(histogram_data, histogram_n_bins, histogram_range)

    if inline_data:
        buffer = io.BytesIO()
        point_data.to_feather(buffer, compression="uncompressed")
        buffer.seek(0)
        arrow_bytes = buffer.read()
        gzipped_bytes = gzip.compress(arrow_bytes)        
        base64_point_data = base64.b64encode(gzipped_bytes).decode()
        json_bytes = json.dumps(hover_data.to_dict(orient="list")).encode()
        gzipped_bytes = gzip.compress(json_bytes)
        base64_hover_data = base64.b64encode(gzipped_bytes).decode()
        label_data_json = label_dataframe.to_json(orient="records")
        gzipped_label_data = gzip.compress(bytes(label_data_json, "utf-8"))
        base64_label_data = base64.b64encode(gzipped_label_data).decode()
        if enable_histogram:
            json_bytes = bin_data.to_json(orient="records", date_format="iso", date_unit='s').encode()
            gzipped_bytes = gzip.compress(json_bytes)
            base64_histogram_bin_data = base64.b64encode(gzipped_bytes).decode()
            buffer = io.BytesIO()
            index_data.to_frame().to_feather(buffer, compression="uncompressed")
            buffer.seek(0)
            arrow_bytes = buffer.read()
            gzipped_bytes = gzip.compress(arrow_bytes)
            base64_histogram_index_data = base64.b64encode(gzipped_bytes).decode()
        else:
            base64_histogram_bin_data = None
            base64_histogram_index_data = None
        file_prefix = None
    else:
        base64_point_data = ""
        base64_hover_data = ""
        base64_label_data = ""
        base64_histogram_bin_data = ""
        base64_histogram_index_data = ""
        file_prefix = (
            offline_data_prefix if offline_data_prefix is not None else "datamapplot"
        )
        with gzip.open(f"{file_prefix}_point_data.zip", "wb") as f:
            point_data.to_feather(f, compression="uncompressed")
        with gzip.open(f"{file_prefix}_meta_data.zip", "wb") as f:
            f.write(json.dumps(hover_data.to_dict(orient="list")).encode())
        label_data_json = label_dataframe.to_json(path_or_buf=None, orient="records")
        with gzip.open(f"{file_prefix}_label_data.zip", "wb") as f:
            f.write(bytes(label_data_json, "utf-8"))
        if enable_histogram:
            with gzip.open(f"{file_prefix}_histogram_bin_data.zip", "wb") as f:
                f.write(bin_data.to_json(orient="records", date_format="iso", date_unit='s').encode())
            with gzip.open(f"{file_prefix}_histogram_index_data.zip", "wb") as f:
                index_data.to_frame().to_feather(f, compression="uncompressed")

    title_font_color = "#000000" if not darkmode else "#ffffff"
    sub_title_font_color = "#777777"
    title_background = "#ffffffaa" if not darkmode else "#000000aa"
    shadow_color = "#aaaaaa44" if not darkmode else "#00000044"
    input_background = "#ffffffdd" if not darkmode else "#000000dd"
    input_border = "#ddddddff" if not darkmode else "222222ff"

    if tooltip_css is None:
        tooltip_css_template = jinja2.Template(_TOOL_TIP_CSS)
        tooltip_css = tooltip_css_template.render(
            title_font_family=tooltip_font_family or font_family,
            title_font_color=title_font_color,
            title_font_weight=tooltip_font_weight,
            title_background=title_background,
            shadow_color=shadow_color,
        )

    if background_color is None:
        page_background_color = "#ffffff" if not darkmode else "#000000"
    else:
        page_background_color = background_color

    # Pepare JS/CSS dependencies for embedding in the HTML template
    dependencies_ctx = {
        "js_dependency_urls": _get_js_dependency_urls(enable_histogram, selection_handler),
        "js_dependency_srcs": _get_js_dependency_sources(
            minify_deps, enable_search, enable_histogram, enable_lasso_selection,
        ),
        "css_dependency_srcs": _get_css_dependency_sources(
            minify_deps, enable_histogram, show_loading_progress
        ),
    }

    template = jinja2.Template(_DECKGL_TEMPLATE_STR)
    api_fontname = font_family.replace(" ", "+")
    font_data = get_google_font_for_embedding(font_family)
    if font_data == "":
        api_fontname = None
    if tooltip_font_family is not None:
        api_tooltip_fontname = tooltip_font_family.replace(" ", "+")
        resp = requests.get(
            f"https://fonts.googleapis.com/css?family={api_tooltip_fontname}",
            timeout=30,
        )
        if not resp.ok:
            api_tooltip_fontname = None
    else:
        api_tooltip_fontname = None

    if selection_handler is not None:
        if custom_html is None:
            custom_html = selection_handler.html
        else:
            custom_html += selection_handler.html

        if custom_js is None:
            custom_js = selection_handler.javascript
        else:
            custom_js += selection_handler.javascript

        if custom_css is None:
            custom_css = selection_handler.css
        else:
            custom_css += selection_handler.css

    html_str = template.render(
        title=title if title is not None else "Interactive Data Map",
        sub_title=sub_title if sub_title is not None else "",
        google_font=api_fontname,
        google_font_data=font_data,
        google_tooltip_font=api_tooltip_fontname,
        page_background_color=page_background_color,
        search=enable_search,
        **histogram_ctx,
        title_font_family=font_family,
        title_font_color=title_font_color,
        title_background=title_background,
        tooltip_css=tooltip_css,
        shadow_color=shadow_color,
        input_background=input_background,
        input_border=input_border,
        custom_css=custom_css,
        use_title=title is not None,
        title_font_size=title_font_size,
        sub_title_font_size=sub_title_font_size,
        sub_title_font_color=sub_title_font_color,
        logo=logo,
        logo_width=logo_width,
        custom_html=custom_html,
        inline_data=inline_data,
        base64_point_data=base64_point_data,
        base64_hover_data=base64_hover_data,
        base64_label_data=base64_label_data,
        base64_histogram_bin_data=base64_histogram_bin_data,
        base64_histogram_index_data=base64_histogram_index_data,
        file_prefix=file_prefix,
        point_size=point_size,
        point_outline_color=point_outline_color,
        point_line_width=point_line_width,
        point_hover_color=[int(c * 255) for c in to_rgba(point_hover_color)],
        point_line_width_max_pixels=point_line_width_max_pixels,
        point_line_width_min_pixels=point_line_width_min_pixels,
        point_radius_max_pixels=point_radius_max_pixels,
        point_radius_min_pixels=point_radius_min_pixels,
        label_text_color=label_text_color,
        line_spacing=line_spacing,
        text_min_pixel_size=text_min_pixel_size,
        text_max_pixel_size=text_max_pixel_size,
        text_outline_width=text_outline_width,
        text_outline_color=[int(c * 255) for c in to_rgba(text_outline_color)],
        text_background_color=text_background_color,
        font_family=font_family,
        font_weight=font_weight,
        text_collision_size_scale=text_collision_size_scale,
        cluster_boundary_polygons="polygon" in label_dataframe.columns,
        cluster_boundary_line_width=cluster_boundary_line_width,
        zoom_level=zoom_level,
        data_center_x=data_center[0],
        data_center_y=data_center[1],
        on_click=on_click,
        enable_lasso_selection=enable_lasso_selection,
        get_tooltip=get_tooltip,
        search_field=search_field,
        show_loading_progress=show_loading_progress,
        custom_js=custom_js,
        **dependencies_ctx,
    )
    return html_str



from sklearn.feature_extraction.text import ENGLISH_STOP_WORDS
import string

_DEFAULT_TAG_COLORS = [
    "#1f77b4", "#ff7f0e","#2ca02c","#d62728","#9467bd","#8c564b","#e377c2","#7f7f7f",
    "#bcbd22","#17becf","#a6008a","#656100","#8aa6ff","#007155","#ce968a","#6139f3",
    "#82b692","#ae8210","#ae9ebe","#5d5d79","#ce86ff","#398e92","#b65504","#ce31d7",
    "#758a55","#9204c6","#187100","#965982","#ef6959","#5d79ff","#7986b2","#b2a66d",
    "#5d614d","#009e71","#00a2e7","#8ea6ae","#8a9a0c","#9e7d5d","#00c66d","#246979",
    "#65c210","#865510","#a23118","#9e7dff","#9239fb","#00c2a6","#ae7d9e","#6165b2",
    "#aaa69e","#005def","#754d8e","#ce7d49","#ba5549","#f35dff","#df9600","#be4dff",
    "#55716d","#8ab65d","#6d9686","#e75500","#75616d","#4d713d","#5d8200","#9e45a6",
    "#7daed7","#867596","#5d798e","#ba75c6","#be55a2","#827135","#008641","#5d96b2",
    "#ae9ae7","#61a261","#b6756d","#5daaa6","#eb41c6","#8e9e7d","#9e8e96","#b69e10",
    "#6d49b6","#867d00","#a66d2d","#ca92c6","#6592df","#4d8265","#7d6d5d","#7d65ef",
    "#45658a","#8a8e9e","#d29a55","#b220df","#9a8e4d","#0086eb","#00829e","#969eca",
    "#c614aa","#007975","#9a86be","#5d6165","#c67100","#755939","#9a4d24","#8e3d7d",
    "#c23900","#6d7961","#eb8a69","#35baeb","#b29679","#718a8e","#9e69a2","#ae75e3",
    "#008a00","#3561ae","#8e9692","#a66549","#7d82db","#00a2b6","#24b682","#9e00aa",
    "#08ba39","#8a49ba","#75659e","#008e79","#5579c6","#927186","#558a41","#755171",
]


class SelectionHandlerBase:
    """Base class for selection handlers. Selection handlers are used to define custom behavior
    when text items are selected on the plot. This can include displaying additional information
    about the selected text items, generating visualizations based on the selected text items, or
    interacting with external APIs to process the selected text items.

    Parameters
    ----------
    dependencies : list, optional
        A list of URLs for external dependencies required by the selection handler. Default is an empty list.

    """

    def __init__(self, **kwargs):
        if "dependencies" in kwargs:
            self.dependencies = kwargs["dependencies"]
        else:
            self.dependencies = []

    @property
    def javascript(self):
        return ""

    @property
    def css(self):
        return ""

    @property
    def html(self):
        return ""


class DisplaySample(SelectionHandlerBase):
    """A selection handler that displays a sample of the selected text items in a container on the page.
    The sample will be randomly selected from the full selection of text items, and the number of samples
    can be controlled by the `n_samples` parameter.

    A resample button is also provided to generate a new sample of text items from the selection.

    Parameters
    ----------
    n_samples : int, optional
        The number of samples to display. Default is 256.

    font_family : str, optional
        The font family to use for the displayed text. Default is None.

    **kwargs
        Additional keyword arguments to pass to the SelectionHandlerBase constructor.

    """

    def __init__(self, n_samples=256, font_family=None, **kwargs):
        super().__init__(
            dependencies=[
                "https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js"
            ],
            **kwargs,
        )
        self.n_samples = n_samples
        self.font_family = font_family

    @property
    def javascript(self):
        return f"""
const resampleButton = document.getElementsByClassName("resample-button")[0]
const clearSelectionButton = document.getElementsByClassName("clear-selection-button")[0]
resampleButton.onclick = resampleSelection
clearSelectionButton.onclick = clearSelection

const shuffle = ([...arr]) => {{
  let m = arr.length;
  while (m) {{
    const i = Math.floor(Math.random() * m--);
    [arr[m], arr[i]] = [arr[i], arr[m]];
  }}
  return arr;
}};
const sampleSize = ([...arr], n = 1) => shuffle(arr).slice(0, n);

function lassoSelectionCallback(selectedPoints) {{
    const n_samples = {self.n_samples};
    if (selectedPoints.length == 0) {{
        const selectionContainer = document.getElementById('selection-container');
        selectionContainer.style.display = 'none';
        return;       
    }}
    if (selectedPoints.length > n_samples) {{
        selectedPoints = sampleSize(selectedPoints, n_samples);
    }}
    const selectionContainer = document.getElementById('selection-container');
    const selectionDisplayDiv = document.getElementById('selection-display');
    var listItems = document.createElement('ul');
    while (selectionDisplayDiv.firstChild) {{
        selectionDisplayDiv.removeChild(selectionDisplayDiv.firstChild);
    }}
    if (datamap.metaData) {{
      selectedPoints.forEach((index) => {{
          listItems.appendChild(document.createElement('li')).textContent = datamap.metaData.hover_text[index];
      }});
    }} else {{
        listItems.appendChild(document.createElement('li')).textContent = "Meta data still loading ..."
    }}
    selectionDisplayDiv.appendChild(listItems);
    $(selectionContainer).animate({{width:'show'}}, 500);
}}

function resampleSelection() {{
    const n_samples = {self.n_samples};
    let selectedPoints = Array.from(datamap.getSelectedIndices());
    if (selectedPoints.length > n_samples) {{
        selectedPoints = sampleSize(selectedPoints, n_samples);
    }}
    const selectionContainer = document.getElementById('selection-container');
    const selectionDisplayDiv = document.getElementById('selection-display');
    var listItems = document.createElement('ul');
    while (selectionDisplayDiv.firstChild) {{
        selectionDisplayDiv.removeChild(selectionDisplayDiv.firstChild);
    }}
    if (datamap.metaData) {{
      selectedPoints.forEach((index) => {{
          listItems.appendChild(document.createElement('li')).textContent = datamap.metaData.hover_text[index];
      }});
    }} else {{
        listItems.appendChild(document.createElement('li')).textContent = "Meta data still loading ..."
    }}
    selectionDisplayDiv.appendChild(listItems);
}}

function clearSelection() {{
    const selectionContainer = document.getElementById('selection-container');
    $(selectionContainer).animate({{width:'hide'}}, 500);

    datamap.removeSelection(datamap.lassoSelectionItemId);
}}
        """

    @property
    def css(self):
        if self.font_family:
            font_family_str = f"font-family: {self.font_family};"
        else:
            font_family_str = ""
        return f"""
    #selection-container {{
        display: none;
        position: absolute;
        top: 0;
        right: 0;
        height: 95%;
        max-width: 33%;
        z-index: 10;
        {font_family_str}
    }}
    #selection-display {{
        overflow-y: auto;
        max-height: 95%;
        margin: 8px;
    }}
    .button {{
        border: none;
        padding: 12px 24px;
        text-align: center;
        display: inline-block;
        margin: 4px 2px;
        cursor: pointer;
        border-radius: 8px;
    }}
    .resample-button {{
        background-color: #4CAF50;
        color: white;
    }}
    .clear-selection-button {{
        position: absolute;
        top: 0;
        right: 0;
        margin: 16px 14px;
        padding: 4px 8px;
        background-color: #b42316;
        color: white;
    }}
    .clear-selection-button:after {{
        font-size: 20px;
        content: "×";
    }}
        """

    @property
    def html(self):
        return f"""
    <div id="selection-container" class="container-box more-opaque">
        <button class="button resample-button">Resample</button>
        <button class="button clear-selection-button"></button>
        <div id="selection-display"></div>
    </div>
        """


class WordCloud(SelectionHandlerBase):
    """A selection handler that generates a word cloud from the selected text items. The word cloud
    is displayed in a container on the page, and the number of words in the cloud can be controlled
    by the `n_words` parameter.

    The word cloud is generated using the d3-cloud library, and the appearance of the word cloud can
    be customized using the `width`, `height`, `font_family`, `stop_words`, `n_rotations`, and `color_scale`
    parameters.

    Parameters
    ----------
    n_words : int, optional
        The number of words to display in the word cloud. Default is 256.

    width : int, optional
        The width of the word cloud container. Default is 500.

    height : int, optional
        The height of the word cloud container. Default is 500.

    font_family : str, optional
        The font family to use for the word cloud. Default is None.

    stop_words : list, optional
        A list of stop words to exclude from the word cloud. Default is the English stop words from scikit-learn.

    n_rotations : int, optional
        The number of rotations to use for the words in the word cloud. Default is 0. More rotations can make the
        word cloud more visually interesting, at the cost of readability.

    color_scale : str, optional
        The color scale to use for the word cloud. Default is "YlGnBu". The color scale can be any d3 color scale
        name, with an optional "_r" suffix to reverse the color scale.

    location : tuple, optional
        The location of the word cloud container on the page. Default is ("bottom", "right").

    **kwargs
        Additional keyword arguments to pass to the SelectionHandlerBase constructor.

    """

    def __init__(
        self,
        n_words=256,
        width=500,
        height=500,
        font_family=None,
        stop_words=None,
        n_rotations=0,
        color_scale="YlGnBu",
        location=("bottom", "right"),
        **kwargs,
    ):
        super().__init__(
            dependencies=[
                "https://d3js.org/d3.v6.min.js",
                "https://unpkg.com/d3-cloud@1.2.7/build/d3.layout.cloud.js",
                "https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js",
            ],
            **kwargs,
        )
        self.n_words = n_words
        self.width = width
        self.height = height
        self.font_family = font_family
        self.stop_words = stop_words or list(ENGLISH_STOP_WORDS)
        self.n_rotations = min(22, n_rotations)
        self.location = location
        if color_scale.endswith("_r"):
            self.color_scale = string.capwords(color_scale[:1]) + color_scale[1:-2]
            self.color_scale_reversed = True
        else:
            self.color_scale = string.capwords(color_scale[:1]) + color_scale[1:]
            self.color_scale_reversed = False

    @property
    def javascript(self):
        return f"""
const _STOPWORDS = new Set({self.stop_words});
const _ROTATIONS = [0, -90, 90, -45, 45, -30, 30, -60, 60, -15, 15, -75, 75, -7.5, 7.5, -22.5, 22.5, -52.5, 52.5, -37.5, 37.5, -67.5, 67.5];
const wordCloudSvg = d3.select("#word-cloud").append("svg")
    .attr("width", {self.width})
    .attr("height", {self.height})
    .append("g")
    .attr("transform", "translate(" + {self.width} / 2 + "," + {self.height} / 2 + ")");
const wordCloudItem = document.getElementById("word-cloud");

function wordCounter(textItems) {{
    const words = textItems.join(' ').toLowerCase().split(/\s+/);
    const wordCounts = new Map();
    words.forEach(word => {{
        wordCounts.set(word, (wordCounts.get(word) || 0) + 1);
    }});
    _STOPWORDS.forEach(stopword => wordCounts.delete(stopword));
    const result = Array.from(wordCounts, ([word, frequency]) => ({{ text: word, size: Math.sqrt(frequency) }}))
                        .sort((a, b) => b.size - a.size).slice(0, {self.n_words});
    const maxSize = Math.max(...(result.map(x => x.size)));
    return result.map(({{text, size}}) => ({{ text: text, size: (size / maxSize)}}));
}}

function generateWordCloud(words) {{
  const width = {self.width};
  const height = {self.height};

  const colorScale = d3.scaleSequential(d3.interpolate{self.color_scale}).domain([{"width / 10, 0" if self.color_scale_reversed else "0, width / 10"}]);

  // Configure a cloud layout
  const layout = d3.layout.cloud()
    .size([width, height])
    .words(words.map(d => ({{text: d.text, size: d.size * width / 10}})))
    .padding(1)
    .rotate(() => _ROTATIONS[~~(Math.random() * {self.n_rotations})])
    .font("{self.font_family or 'Impact'}")
    .fontSize(d => d.size)
    .fontWeight(d => Math.max(300, Math.min(d.size * 9000 / width, 900)))
    .on("end", draw);

  layout.start();

  function draw(words) {{
    const t = d3.transition().duration(500);
    
    // Update existing words
    const text = wordCloudSvg.selectAll("text")
      .data(words, d => d.text);
    
    // Remove old words
    text.exit()
      .transition(t)
      .attr("fill-opacity", 0)
      .attr("font-size", 1)
      .remove();
    // Add new words
    text.enter()
      .append("text")
      .attr("text-anchor", "middle")
      .attr("fill-opacity", 0)
      .attr("font-size", 1)
      .attr("font-family", "{self.font_family or 'Impact'}")
      .text(d => d.text)
      .merge(text) // Merge enter and update selections
      .transition(t)
      .attr("transform", d => "translate(" + [d.x, d.y] + ")rotate(" + d.rotate + ")")
      .attr("fill-opacity", 1)
      .attr("font-size", d => d.size)
      .attr("font-weight", d => Math.max(300, Math.min(d.size * 9000 / width, 900)))
      .attr("fill", d => colorScale(d.size));
  }}
}}

function lassoSelectionCallback(selectedPoints) {{
    if (selectedPoints.length > 0) {{
        $(wordCloudItem).animate({{height:'show'}}, 250);
    }} else {{
        $(wordCloudItem).animate({{height:'hide'}}, 250);
    }}
    let selectedText;
    if (datamap.metaData) {{
        selectedText = selectedPoints.map(i => datamap.metaData.hover_text[i]);
    }} else {{
        selectedText = ["Meta data still loading ..."];
    }}
    const wordCounts = wordCounter(selectedText);
    generateWordCloud(wordCounts);
}}
"""

    @property
    def html(self):
        return """<div id="word-cloud" class="container-box"></div>"""

    @property
    def css(self):
        return f"""
#word-cloud {{
    position: absolute;
    {self.location[1]}: 0;
    {self.location[0]}: 0;
    display: none;
    width: {self.width}px;
    height: {self.height}px;
    z-index: 10;
}}
"""


class CohereSummary(SelectionHandlerBase):
    """A selection handler that uses the Cohere API to generate a summary of selected text items.
    The handler requires an API key to be provided by the end-user in the resulting HTML page.
    The handler will generate a prompt based on the selected text items and keywords extracted
    from the text items, and get a Cohere model to summarize this. The summary will be displayed
    in a container on the page.

    This handler can likely be adapted to other API services that provide text summarization.

    Note that the API key handling here is secure enough for private or small-scale use,
    but is not suitable for enterprise or production use.

    Parameters
    ----------
    model : str, optional
        The Cohere model to use for summarization. Default is "command-r". See the Cohere docs
        for more information on available models.

    stop_words : list, optional
        A list of stop words to exclude from the keyword extraction. Default is the English stop
        words from scikit-learn.

    n_keywords : int, optional
        The number of keywords to extract from the text items. Default is 128.

    n_samples : int, optional
        The number of samples to use for the summary. Default is 64.

    width : int, optional
        The width of the summary container. Default is 500.

    location : tuple, optional
        The location of the summary container on the page. Default is ("top", "right").

    **kwargs
        Additional keyword arguments to pass to the SelectionHandlerBase constructor.
    """

    def __init__(
        self,
        model="command-r",
        stop_words=None,
        n_keywords=128,
        n_samples=64,
        width=500,
        location=("top", "right"),
        **kwargs,
    ):
        super().__init__(
            dependencies=[
                "https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js",
            ],
            **kwargs,
        )
        self.model = model
        self.stop_words = stop_words or list(ENGLISH_STOP_WORDS)
        self.n_keywords = n_keywords
        self.n_samples = n_samples
        self.width = width
        self.location = location

    @property
    def javascript(self):
        return f"""
// Stop word list
const _STOPWORDS = new Set({self.stop_words});
const summaryContainer = document.getElementById('summary-container');

// Cohere API call
async function cohereChat(message, apiKey) {{
  const response = await fetch('https://api.cohere.ai/v1/chat', {{
    method: 'POST',
    headers: {{
      'Authorization': `Bearer ${{apiKey}}`,
      'Content-Type': 'application/json'
    }},
    body: JSON.stringify({{
      message: message,
      model: "{self.model}"
    }})
  }});
  if (!response.ok) {{
    if (response.status === 401) {{
        return {{ text: "Error! Unauthorized: Please check your API key." }};
    }} else {{
        return {{ text: `Error! status: ${{response.status}}` }};
    }}
  }}
  return await response.json();
}}

// Word counts for keywords
function wordCounter(textItems) {{
    const words = textItems.join(' ').toLowerCase().split(/\s+/);
    const wordCounts = new Map();
    words.forEach(word => {{
        wordCounts.set(word, (wordCounts.get(word) || 0) + 1);
    }});
    _STOPWORDS.forEach(stopword => wordCounts.delete(stopword));
    const result = Array.from(wordCounts, ([word, frequency]) => ({{ text: word, size: Math.sqrt(frequency) }}))
                        .sort((a, b) => b.size - a.size).slice(0, {self.n_keywords});
    const maxSize = Math.max(...(result.map(x => x.size)));
    return result.map(({{text, size}}) => ({{ text: text, size: (size / maxSize)}}));
}}

// Array shuffling for random sampling
const shuffle = ([...arr]) => {{
  let m = arr.length;
  while (m) {{
    const i = Math.floor(Math.random() * m--);
    [arr[m], arr[i]] = [arr[i], arr[m]];
  }}
  return arr;
}};
const sampleSize = ([...arr], n = 1) => shuffle(arr).slice(0, n);

// Create a summary using Cohere API
function generateSummary(textItems) {{
    const apiKey = document.getElementById("api-key").value;
    if (apiKey === "") {{
      summaryContainer.innerHTML = "No API Key provided ... cannot generate a summary!"
    }} else {{
        const keywords = wordCounter(textItems).map(d => d.text);
        const sample_text = sampleSize(textItems, {self.n_samples});

        // Build prompt from keywords and samples with some framing text
        const prompt = `We have samples of items from a selection of items about a topic.
Keywords associated to the topic are: ${{keywords.join(", ")}}
Samples text items associated to the topic are: 
- ${{sample_text.join("\\n  - ")}}

Please provide a concise summary of the selection of items. Be as specific as possible.
The summary should be a few sentences long at most, and ideally just a single sentence.
`;
        cohereChat(prompt, apiKey).then(response => {{ summaryContainer.innerHTML = response.text }} );
    }}
}}

function lassoSelectionCallback(selectedPoints) {{
    if (selectedPoints.length > 0) {{
        $(summaryContainer).animate({{width:'show'}}, {self.width});
    }} else {{
        $(summaryContainer).animate({{width:'hide'}}, {self.width});
    }}
    let selectedText;
    if (datamap.metaData) {{
        selectedText = selectedPoints.map(i => datamap.metaData.hover_text[i]);
    }} else {{
        selectedText = ["Meta data still loading ..."];
    }}
    generateSummary(selectedText);
}}
        """

    @property
    def html(self):
        return """
      <div id="api-key-container" class="container-box">
        <label for="apiKey">Cohere API Key: </label>
        <input autocomplete="off" type="password" id="api-key" placeholder="Enter your API key here" />
      </div> 
      <div id="summary-container" class="container-box more-opaque"></div>
"""

    @property
    def css(self):
        return f"""
#summary-container {{
    position: absolute;
    {self.location[1]}: 0;
    {self.location[0]}: 5%;
    display: none;
    width: {self.width}px;
    height: fit-content;
    font-size: 16px;
    font-weight: 400;
    z-index: 10;
}}
#api-key-container {{
    position: absolute;
    {self.location[1]}: 0;
    {self.location[0]}: 0;
    width: fit-content;
    z-index: 10;
}}
"""


class TagSelection(SelectionHandlerBase):
    """
    A selection handler that allows users to create and save tags for selected items.
    The handler provides a container for displaying existing tags, a button to create a new tag,
    and a button to save the tags to a JSON file.

    The handler also provides a visual indicator for selected items that have been tagged, and
    allows users to add selected items to existing tags.

    Parameters
    ----------
    tag_colors : list, optional
        A list of colors to use for the tags. Default is a set of default colors extending the tab10 palette.
    """

    def __init__(self, tag_colors=None, **kwargs):
        super().__init__(**kwargs)
        if tag_colors is None:
            self.tag_colors = _DEFAULT_TAG_COLORS
        else:
            self.tag_colors = tag_colors

    @property
    def javascript(self):
        return f"""
    const tagColors = [
    {",".join(['"'+x+'"' for x in self.tag_colors])}
    ];

    const tags = new Map();
    const tagButton = document.getElementById("new-tag-button");
    const tagList = document.getElementById("tag-list");
    const tagInput = document.getElementById("tag-input");
    const saveTagsButton = document.getElementById("save-tags");
    const selectedTags = new Set();
    saveTagsButton.onclick = saveTags;
    let numTags = 0;
    
    function mapToObject(map) {{
        const obj = {{}};
        map.forEach((value, key) => {{
            obj[key] = Array.from(value);
        }});
        return obj;
    }}

    function saveTags() {{
        const tagsObject = mapToObject(tags);
        const jsonString = JSON.stringify(tagsObject, null, 2);
        const blob = new Blob([jsonString], {{ type: 'application/json' }});
        const url = URL.createObjectURL(blob);

        const a = document.createElement('a');
        a.href = url;
        a.download = 'tags.json';  // Specify the filename
        document.body.appendChild(a);
        a.click();
        URL.revokeObjectURL(url);
        document.body.removeChild(a);
    }}

    function toggleTagSelection(tagName) {{
        if (selectedTags.has(tagName)) {{
            selectedTags.delete(tagName);
            const tagSwatch = document.getElementById(`tag-selector-${{tagName}}`);
            tagSwatch.innerHTML = "";
        }} else {{
            selectedTags.add(tagName);
            const tagSwatch = document.getElementById(`tag-selector-${{tagName}}`);
            tagSwatch.innerHTML = "✓";
        }}
        const selectedIndices = [];
        selectedTags.forEach(tag => {{
            tags.get(tag).forEach(index => {{
                selectedIndices.push(index);
            }});
        }});
        datamap.addSelection(selectedIndices, "tag-selection");
    }}

    function addSelectionToTag(tagName, selectedPoints) {{
        tags.set(tagName, tags.get(tagName).union(new Set(selectedPoints)));
        const addToTagButton = document.getElementById(`add-to-${{tagName}}`);
        addToTagButton.classList.remove("enabled");
        addToTagButton.style.display = "none";
        addToTagButton.disable = true;
    }}
    
    function createNewTag(selectedPoints) {{
        const tagName = tagInput.value;
        if (tags.has(tagName)) {{
            alert("Tag already exists! Try adding to the existing tag instead.");
        }} else if (tagName === "") {{
            alert("Tag name cannot be empty!");
        }} else {{
            addTag(tagName);
            tags.set(tagName, new Set(selectedPoints));
            tagInput.value = "";
        }}
    }}

    function addTag(tagName) {{
        const tagItem = document.createElement('li');
        tagItem.innerHTML = `
<div class="row">
  <div class="tag-info">
    <div 
        id="tag-selector-${{tagName}}" 
        class="box" 
        style="background-color: ${{tagColors[numTags]}};" 
    ></div>
    ${{tagName}}
  </div>
  <button id="add-to-${{tagName}}" class="button tag-button add-to-tag-button" style="display: none;">Add to tag</button>
</div>`;
        numTags += 1;
        tagList.appendChild(tagItem);
        document.getElementById(`tag-selector-${{tagName}}`).addEventListener("click", () => toggleTagSelection(tagName));
    }}

    function lassoSelectionCallback(selectedPoints) {{
        if (selectedPoints.length !== 0) {{
            tagButton.classList.add("enabled");
            tagButton.onclick = () => createNewTag(selectedPoints);
            tagButton.disabled = false;
            tags.forEach((points, tagName, map) => {{
                    const addToTagButton = document.getElementById(`add-to-${{tagName}}`);
                    addToTagButton.onclick = () => addSelectionToTag(tagName, selectedPoints);
                    addToTagButton.classList.add("enabled");
                    addToTagButton.style.display = "block";
                    addToTagButton.disable = false;
            }});
        }} else {{
            tagButton.classList.remove("enabled");
            tagButton.disabled = true;
            tags.forEach((points, tagName, map) => {{
                    const addToTagButton = document.getElementById(`add-to-${{tagName}}`);
                    addToTagButton.classList.remove("enabled");
                    addToTagButton.style.display = "none";
                    addToTagButton.disable = true;
            }});  
        }}
    }}
        """

    @property
    def html(self):
        return f"""
    <div id="tag-container" class="container-box more-opaque">
        <div id="tag-display">
            <h3>Existing Tags</h3>
            <ul id="tag-list">
            </ul>
        </div>
        <span>
            <button id="new-tag-button" class="button tag-button">Create New Tag</button>
            <input type="text" id="tag-input" placeholder="Enter tag name">
        </span>
        <button id="save-tags" class="button tag-button enabled">Save tags</button>
    </div>
"""

    @property
    def css(self):
        return f"""
#tag-container {{
    position: absolute;
    top: 0;
    right: 0;
    width: 25%;
    height: fit-content;
    z-index: 10;
}}
#tag-display {{
    overflow-y: auto;
    max-height: 95%;
    margin: 8px;
}}
.tag-button {{
    border: none;
    padding: 4px 8px;
    text-align: center; 
    display: inline-block;
    margin: 4px 2px;
    cursor: pointer;
    border-radius: 8px;
}}
.tag-button:enabled {{
    background-color: #3ba5e7;
    color: white;
}}
.tag-button:disabled {{
    background-color: #cccccc;
    color: #666666;
}}
#save-tags {{
  float: right;
}}
#tag-list {{
    list-style-type: none;
    width: 75%;
}}
.row {{
    display : flex;
    align-items : center;
    width: 100%;
    justify-content: space-between;
}}
.box {{
    height:10px;
    width:10px;
    border-radius:2px;
    margin-right:5px;
    padding:0px 0 1px 0;
    text-align:center;
    color: white;
    font-size: 14px;
}}
.tag-info {{
  display: flex;
  align-items: center;
}}
.add-to-tag-button {{
  float: right;
  font-size: 8px;
  padding: 2px 4px;
  margin: 0px 16px;
}}
    """



import numpy as np
import numba


@numba.njit(
    [
        "f4(f4[::1],f4[::1])",
        numba.types.float32(
            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
            numba.types.Array(numba.types.float32, 1, "C", readonly=True),
        ),
    ],
    fastmath=True,
    locals={
        "result": numba.types.float32,
        "diff": numba.types.float32,
        "dim": numba.types.intp,
        "i": numba.types.uint16,
    },
)
def euclidean(x, y):
    r"""Squared euclidean distance.

    .. math::
        D(x, y) = \sum_i (x_i - y_i)^2
    """
    result = 0.0
    dim = x.shape[0]
    for i in range(dim):
        diff = x[i] - y[i]
        result += diff * diff

    return np.sqrt(result)


@numba.njit(parallel=True, nogil=True)
def chunked_parallel_pairwise_distances(X, Y=None, metric=euclidean, chunk_size=16):
    if Y is None:
        XX, symmetrical = X, True
        row_size = col_size = X.shape[0]
    else:
        XX, symmetrical = Y, False
        row_size, col_size = X.shape[0], Y.shape[0]

    result = np.zeros((row_size, col_size), dtype=np.float32)
    n_row_chunks = (row_size // chunk_size) + 1
    for chunk_idx in numba.prange(n_row_chunks):
        n = chunk_idx * chunk_size
        chunk_end_n = min(n + chunk_size, row_size)
        m_start = n if symmetrical else 0
        for m in range(m_start, col_size, chunk_size):
            chunk_end_m = min(m + chunk_size, col_size)
            for i in range(n, chunk_end_n):
                for j in range(m, chunk_end_m):
                    result[i, j] = metric(X[i], XX[j])
    return result


@numba.njit()
def pull_arms(data, arms, num_pulls_per_arm, estimates, pull_counts):
    other_candidates = np.random.choice(
        data.shape[0], size=num_pulls_per_arm, replace=False
    ).astype(np.int32)
    data_arm = data[arms]
    data_other = data[other_candidates]

    distance_sums = np.sum(
        chunked_parallel_pairwise_distances(data_arm, data_other), axis=1
    )

    estimates *= pull_counts
    estimates += distance_sums
    pull_counts += num_pulls_per_arm
    estimates /= pull_counts


@numba.njit()
def medoid(data, arm_budget=20):
    pull_counts = np.zeros(data.shape[0], dtype=np.int32)
    pull_budget = arm_budget * data.shape[0]
    estimates = np.zeros(data.shape[0], dtype=np.float32)
    current_active_arms = np.arange(data.shape[0])
    n_rounds = int(np.ceil(np.log2(data.shape[0])))

    while current_active_arms.shape[0] > 1:
        num_pulls_per_arm = max(
            1,
            int(
                min(
                    data.shape[0],
                    np.floor(pull_budget / (current_active_arms.shape[0] * n_rounds)),
                )
            ),
        )
        pull_arms(data, current_active_arms, num_pulls_per_arm, estimates, pull_counts)

        median = np.median(estimates)
        mask = estimates <= median
        current_active_arms = current_active_arms[mask]
        estimates = estimates[mask]
        pull_counts = pull_counts[mask]

    return data[current_active_arms[0]]



import numpy as np
import io
from matplotlib import pyplot as plt

try:
    from matplotlib.backend_bases import _get_renderer as matplot_get_renderer
except ImportError:
    matplot_get_renderer = None


def ccw(a, b, c):
    return (c[1] - a[1]) * (b[0] - a[0]) > (b[1] - a[1]) * (c[0] - a[0])


def intersect(a, b, c, d):
    return ccw(a, c, d) != ccw(b, c, d) and ccw(a, b, c) != ccw(a, b, d)


# From bioframe (https://github.com/open2c/bioframe)
def arange_multi(starts, stops):
    lengths = stops - starts

    if np.isscalar(starts):
        starts = np.full(len(stops), starts)
    cat_start = np.repeat(starts, lengths)
    cat_counter = np.arange(lengths.sum()) - np.repeat(
        lengths.cumsum() - lengths, lengths
    )
    cat_range = cat_start + cat_counter
    return cat_range


# From bioframe (https://github.com/open2c/bioframe)
def overlap_intervals(starts1, ends1, starts2, ends2, closed=False, sort=False):
    starts1 = np.asarray(starts1)
    ends1 = np.asarray(ends1)
    starts2 = np.asarray(starts2)
    ends2 = np.asarray(ends2)

    # Concatenate intervals lists
    n1 = len(starts1)
    n2 = len(starts2)
    ids1 = np.arange(0, n1)
    ids2 = np.arange(0, n2)

    # Sort all intervals together
    order1 = np.lexsort([ends1, starts1])
    order2 = np.lexsort([ends2, starts2])
    starts1, ends1, ids1 = starts1[order1], ends1[order1], ids1[order1]
    starts2, ends2, ids2 = starts2[order2], ends2[order2], ids2[order2]

    # Find interval overlaps
    match_2in1_starts = np.searchsorted(starts2, starts1, "left")
    match_2in1_ends = np.searchsorted(starts2, ends1, "right" if closed else "left")
    # "right" is intentional here to avoid duplication
    match_1in2_starts = np.searchsorted(starts1, starts2, "right")
    match_1in2_ends = np.searchsorted(starts1, ends2, "right" if closed else "left")

    # Ignore self-overlaps
    match_2in1_mask = match_2in1_ends > match_2in1_starts
    match_1in2_mask = match_1in2_ends > match_1in2_starts
    match_2in1_starts, match_2in1_ends = (
        match_2in1_starts[match_2in1_mask],
        match_2in1_ends[match_2in1_mask],
    )
    match_1in2_starts, match_1in2_ends = (
        match_1in2_starts[match_1in2_mask],
        match_1in2_ends[match_1in2_mask],
    )

    # Generate IDs of pairs of overlapping intervals
    overlap_ids = np.block(
        [
            [
                np.repeat(ids1[match_2in1_mask], match_2in1_ends - match_2in1_starts)[
                    :, None
                ],
                ids2[arange_multi(match_2in1_starts, match_2in1_ends)][:, None],
            ],
            [
                ids1[arange_multi(match_1in2_starts, match_1in2_ends)][:, None],
                np.repeat(ids2[match_1in2_mask], match_1in2_ends - match_1in2_starts)[
                    :, None
                ],
            ],
        ]
    )

    if sort:
        # Sort overlaps according to the 1st
        overlap_ids = overlap_ids[np.lexsort([overlap_ids[:, 1], overlap_ids[:, 0]])]

    return overlap_ids


# From adjustText (https://github.com/Phyla/adjustText)
def get_renderer(fig):
    # If the backend support get_renderer() or renderer, use that.
    if hasattr(fig.canvas, "get_renderer"):
        return fig.canvas.get_renderer()

    if hasattr(fig.canvas, "renderer"):
        return fig.canvas.renderer

    # Otherwise, if we have the matplotlib function available, use that.
    if matplot_get_renderer:
        return matplot_get_renderer(fig)

    # No dice, try and guess.
    # Write the figure to a temp location, and then retrieve whichever
    # render was used (doesn't work in all matplotlib versions).
    fig.canvas.print_figure(io.BytesIO())
    try:
        return fig._cachedRenderer

    except AttributeError:
        # No luck.
        # We're out of options.
        raise ValueError("Unable to determine renderer") from None


# From adjustText (https://github.com/Phyla/adjustText)
def get_bboxes(objs, r=None, expand=(1, 1), ax=None):
    ax = ax or plt.gca()
    r = r or get_renderer(ax.get_figure())
    return [i.get_window_extent(r).expanded(*expand) for i in objs]


# From adjustText (https://github.com/Phyla/adjustText)
def get_2d_coordinates(objs, expand=(1.0, 1.0)):
    try:
        ax = objs[0].axes
    except:
        ax = objs.axes
    bboxes = get_bboxes(objs, get_renderer(ax.get_figure()), expand, ax)
    xs = [
        (ax.convert_xunits(bbox.xmin), ax.convert_yunits(bbox.xmax)) for bbox in bboxes
    ]
    ys = [
        (ax.convert_xunits(bbox.ymin), ax.convert_yunits(bbox.ymax)) for bbox in bboxes
    ]
    coords = np.hstack([np.array(xs), np.array(ys)])
    return coords


def text_line_overlaps(text_locations, label_locations, text_bounding_boxes):
    result = []
    for i, box in enumerate(text_bounding_boxes):
        for j in range(text_locations.shape[0]):
            if i == j:
                continue

            if intersect(
                text_locations[j], label_locations[j], box[[0, 2]], box[[1, 3]]
            ) or intersect(
                text_locations[j], label_locations[j], box[[0, 3]], box[[1, 2]]
            ):
                result.append((i, j))

    return result



import numpy as np
import pandas as pd

import datashader as ds
import datashader.transfer_functions as tf
from datashader.mpl_ext import dsshow

from functools import partial

from sklearn.neighbors import KernelDensity
from skimage.transform import rescale

from matplotlib import pyplot as plt
from matplotlib import font_manager
from matplotlib import patheffects

from datamapplot.overlap_computations import get_2d_coordinates
from datamapplot.text_placement import (
    estimate_dynamic_font_size,
    initial_text_location_placement,
    fix_crossings,
    adjust_text_locations,
    estimate_font_size,
    pylabeladjust_text_locations,
)

from warnings import warn
from tempfile import NamedTemporaryFile

import requests
import re


class GoogleAPIUnreachable(Warning):
    pass


def _can_reach_google_fonts(timeout: float = 5.0) -> bool:
    try:
        response = requests.get("https://fonts.googleapis.com/css?family=Roboto", timeout=timeout)
        return response.ok
    except requests.RequestException:
        return False


def get_google_font(fontname):
    try:
        api_fontname = fontname.replace(" ", "+")
        api_response = requests.get(
            f"https://fonts.googleapis.com/css?family={api_fontname}:black,bold,regular,light"
        )
        if api_response.ok:
            font_urls = re.findall(r"(https?://[^\)]+)", str(api_response.content))
            for font_url in font_urls:
                font_data = requests.get(font_url)
                f = NamedTemporaryFile(delete=False, suffix=".ttf")
                f.write(font_data.content)
                f.close()
                font_manager.fontManager.addfont(f.name)
    except:
        warn(f"Failed in getting google-font {fontname}; using fallback ...")


def datashader_scatterplot(
    data_map_coords,
    color_list,
    point_size,
    ax,
):
    data = pd.DataFrame(
        {
            "x": data_map_coords.T[0],
            "y": data_map_coords.T[1],
            "label": pd.Categorical(color_list),
        }
    )
    color_key = {x: x for x in np.unique(color_list)}
    dsshow(
        data,
        ds.Point("x", "y"),
        ds.count_cat("label"),
        color_key=color_key,
        norm="eq_hist",
        ax=ax,
        shade_hook=partial(tf.spread, px=point_size, how="over"),
    )
    return ax


def add_glow_to_scatterplot(
    data_map_coords,
    color_list,
    ax,
    noise_color="#999999",
    kernel_bandwidth=0.25,
    approx_patch_size=32,
    kernel="gaussian",
    n_levels=8,
    max_alpha=0.5,
):
    # we are assuming colors are hex strings!
    unique_colors = np.unique(color_list)
    color_array = np.asarray(color_list)

    for color in unique_colors:
        if color == noise_color:
            continue

        cluster_embedding = data_map_coords[color_array == color]

        # find bounds for the cluster
        xmin, xmax = (
            np.min(cluster_embedding.T[0]) - 8 * kernel_bandwidth,
            np.max(cluster_embedding.T[0]) + 8 * kernel_bandwidth,
        )
        ymin, ymax = (
            np.min(cluster_embedding.T[1]) - 8 * kernel_bandwidth,
            np.max(cluster_embedding.T[1]) + 8 * kernel_bandwidth,
        )
        width = xmax - xmin
        height = ymax - ymin
        aspect_ratio = width / height

        # Make an appropriately sized image patch
        patch_size = min(
            max(max(width, height) * approx_patch_size / 6.0, approx_patch_size), 384
        )
        patch_width = int(patch_size * aspect_ratio)
        patch_height = int(patch_size)

        # Build a meshgrid over which to evaluate the KDE
        xs = np.linspace(xmin, xmax, patch_width)
        ys = np.linspace(ymin, ymax, patch_height)
        xv, yv = np.meshgrid(xs, ys[::-1])
        for_scoring = np.vstack([xv.ravel(), yv.ravel()]).T

        # Build the KDE of the cluster
        class_kde = KernelDensity(
            bandwidth=kernel_bandwidth, kernel=kernel, atol=1e-8, rtol=1e-4
        ).fit(cluster_embedding)
        zv = class_kde.score_samples(for_scoring).reshape(xv.shape)
        zv = rescale(zv, 12)
        # Construct colours of varying alpha values for different levels
        alphas = [
            f"{x:02X}"
            for x in np.linspace(0, max_alpha * 255, n_levels, endpoint=True).astype(
                np.uint8
            )
        ]
        level_colors = [color + alpha for alpha in alphas]

        # Create a countour plot for the image patch
        contour_data = np.exp(zv)
        ax.contourf(
            contour_data,
            levels=n_levels,
            colors=level_colors,
            extent=(xmin, xmax, ymin, ymax),
            extend="max",
            origin="upper",
            antialiased=True,
            zorder=0,
        )


def render_plot(
    data_map_coords,
    color_list,
    label_text,
    label_locations,
    label_cluster_sizes,
    *,
    title=None,
    sub_title=None,
    figsize=(12, 12),
    dynamic_label_size=False,
    dynamic_label_size_scaling_factor=0.75,
    font_family="Roboto",
    font_weight=400,
    label_linespacing=0.95,
    label_font_size=None,
    label_text_colors=None,
    label_arrow_colors=None,
    highlight_colors=None,
    point_size=1,
    alpha=1.0,
    dpi=plt.rcParams["figure.dpi"],
    label_over_points=False,
    label_base_radius=None,
    label_margin_factor=1.5,
    min_font_size=4.0,
    max_font_size=24.0,
    min_font_weight=200,
    max_font_weight=800,
    highlight_labels=None,
    highlight_label_keywords={"fontweight": 1000},
    add_glow=True,
    noise_color="#999999",
    glow_keywords={
        "kernel": "gaussian",
        "kernel_bandwidth": 0.25,
        "approx_patch_size": 64,
    },
    darkmode=False,
    logo=None,
    logo_width=0.15,
    force_matplotlib=False,
    label_direction_bias=None,
    marker_type="o",
    marker_size_array=None,
    arrowprops={},
    title_keywords=None,
    sub_title_keywords=None,
    pylabeladjust_speed=None,
    pylabeladjust_max_iterations=500,
    pylabeladjust_adjust_by_size=True,
    pylabeladjust_margin_percentage=7.5,
    pylabeladjust_radius_scale=1.05,
    label_font_stroke_width=3,
    label_font_outline_alpha=0.5,
    ax=None,
    verbose=False,
):
    """Render a static data map plot with given colours and label locations and text. This is
    a lower level function, and should usually not be used directly unless there are specific
    reasons for digging in. This usually involves things like getting direct control over label
    locations, altering label texts to suit specific needs, or direct control over point
    colouring in the scatterplot.

    All keyword arguments from ``create_plot`` are passed on to ``render_plot``, so any
    *keyword* arguments here are also valid keyword arguments for ``create_plot``.

    Parameters
    ----------
    data_map_coords: ndarray of floats of shape (n_samples, 2)
        The 2D coordinates for the data map. Usually this is produced via a
        dimension reduction technique such as UMAP, t-SNE, PacMAP, PyMDE etc.

    color_list: iterable of str of len n_samples
        A list of hex-string colours, one per sample, for colouring points in the
        scatterplot of the data map.

    label_text: list of str
        A list of label text strings, one per unique label.

    label_locations: ndarray of floats of shape (n_labels, 2)
        An array of the "location" (usually centroid) of the cluster of the
        associated text label (see ``label_text``).

    title: str or None (optional, default=None)
        A title for the plot. If ``None`` then no title is used for the plot.
        The title should be succint; three to seven words.

    sub_title: str or None (optional, default=None)
        A sub-title for the plot. If ``None`` then no sub-title is used for the plot.
        The sub-title can be significantly longer then the title and provide more information\
        about the plot and data sources.

    figsize: (int, int) (optional, default=(12,12))
        How big to make the figure in inches (actual pixel size will depend on ``dpi``).

    dynamic_label_size: bool (optional, default=False)
        Whether to use dynamic label sizing based on the sizes of the clusters.

    dynamic_label_size_scaling_factor: float (optional, default=0.75)
        The scaling factor to use when using dynamic label sizing based on the sizes of the clusters.

    font_family: str (optional, default="DejaVu Sans")
        The font_family to use for the plot -- the labels and the title and sub-title
        unless explicitly over-ridden by title_keywords or sub_title_keywords.

    label_linespacing: float (optional, default=0.95)
        The line-spacing to use when rendering multi-line labels in the plot. The default
        of 0.95 keeps multi-line labels compact, but can be less than ideal for some fonts.

    label_font_size: float or None (optional, default=None)
        The font-size (in pts) to use for the text labels in the plot. If this is ``None``
        then a heuristic will be used to try to find the best font size that can fit all
        the labels in.

    label_text_colors: str or list of str or None (optional, default=None)
        The colours of the text labels, one per text label. If None then the text labels
        will be either black or white depending on ``darkmode``. If just a single string
        then it is assumed to be a fixed colour for all labels.

    label_arrow_colors: str or list of str or None (optional, default=None)
        The colours of the arrows between the text labels and clusters, one per text label.
        If None then the arrows will be either black or white depending on ``darkmode``.
        If just a single string then it is assumed to be a fixed colour for all arrows.

    highlight_colors: list of str or None (optional default=None)
        The colours used if text labels are highlighted and a bounding box around the label is
        used. For example ``create_plot`` uses the cluster colours from the colour mapping that
        was passed or created.

    point_size: int or float (optional, default=1)
        How big to make points in the scatterplot rendering of the data map. Depending on
        whether you are in datashader mode or matplotlib mode this can either be an
        int (datashader) or a float (matplotlib). If in datashader mode this is explicitly
        the radius, in number of pixels, that each point should be. If in matplotlib mode
        this is the matplotlib scatterplot size, which can be relative to the plot-size
         and other factors.

    alpha: float (optional, default=1.0)
        The alpha transparency value to use when rendering points.

    dpi: int (optional, default=plt.rcParams["figure.dpi"])
        The dots-per-inch to use when rendering the plot.

    label_over_points: bool (optional, default=False)
        Whether to attempt tom place text labels directly on top of the points in clusters. This
        can result in severe over-packing, and this is remedied via pylabeladjust which can end up
        moving labels some distance. For smaller numbers of labels this is likely a good choice, for
        more than 20 labels this will require a small font. For larger numbers of labels still this
        may be sub-optimal.

    label_base_radius: float or None (optional, default=None)
        Labels are placed in rings around the data map. This value can explicitly control the
        radius (in data coordinates) of the innermost such ring.

    label_margin_factor: float (optional, default=1.5)
        The expansion factor to use when creating a bounding box around the label text
        to compute whether overlaps are occurring during the label placement adjustment phase.

    min_font_size: float (optional, default=4.0)
        The minimum font size to use when estimating the font size for the labels.

    max_font_size: float (optional, default=24.0)
        The maximum font size to use when estimating the font size for the labels.

    min_font_weight: int (optional, default=200)
        The minimum font weight to use when using dynamic label sizing (font weights will vary as well).

    max_font_weight: int (optional, default=800)
        The maximum font weight to use when using dynamic label sizing (font weights will vary as well).

    highlight_labels: list of str or None (optional, default=None)
        A list of the labels to be highlighted.

    highlight_label_keywords: dict (optional, default={"fontweight": "bold"})
        Keywords for how to highlight the labels. This dict will be passed on as keyword
        arguments to the matplotlib ``annotate`` function. See the matplotlib documentation
        for more details on what can be done.

    add_glow: bool (optional, default=True)
        Whether to add a glow-effect using KDEs.

    noise_color: str (optional, default="#999999")
        The colour to use for unlabelled or noise points in the data map. This should usually
        be a muted or neutral colour to distinguish background points from the labelled clusters.

    glow_keywords: dict (optional, default={"kernel": "gaussian","kernel_bandwidth": 0.25})
        Keyword arguments that will be passed along to the ``add_glow_to_scatterplot``
        function. See that function for more details.

    darkmode: bool (optional, default=False)
        Whether to render the plot in darkmode (with a dark background) or not.

    logo: ndarray or None (optional, default=None)
        A numpy array representation of an image (suitable for matplotlib's ``imshow``) to
        be used as a logo placed in the bottom right corner of the plot.

    logo_width: float (optional, default=0.15)
        The width, as a fraction of the total figure width, of the logo.

    force_matplotlib: bool (optional, default=False)
        Force using matplotlib instead of datashader for rendering the scatterplot of the
        data map. This can be useful if you wish to have a different marker_type, or variably
        sized markers based on a marker_size_array, neither of which are supported by the
        datashader based renderer.

    label_direction_bias: float or None (optional, default=None)
        When placing labels in rings, how much bias to place toward east-west compass points
        as opposed to north-south. A value of 1.0 provides no bias (uniform placement around
        the circle). Values larger than one will place more labels ion the east-west areas.

    marker_type: str (optional, default="o")
        The type of marker to use for rendering the scatterplot. This is only valid if
        matplotlib mode is being used. Valid marker_types are any matplotlib marker string.
        See the matplotlib marker documentation for more details.

    marker_size_array: ndarray of shape (n_samples,) or None (optional, default=None)
        The (variable) size or markers to use. This is only valid if matplotlib mode is being
        used. This should be an array of (matplotlib) marker sizes as you would use for the
        ``s`` argument in ``matplotlib.pyplot.scatterplot``.

    arrowprops: dict (optional default={})
        A dict of keyword argumetns to pass through to the ``arrowprops`` argument of
        ``matplotlib.pyplot.annotate``. This allows for control of arrow-styles,
        connection-styles, linewidths, colours etc. See the documentation of matplotlib's
        ``annotate`` function for more details.

    title_keywords: dict or None (optional, default=None)
        A dictionary of keyword arguments to pass through to matplotlib's ``suptitle`` fucntion.
        This includes things like fontfamily, fontsize, fontweight, color, etc.

    sub_title_keywords: dict or None (optional, default=None)
        A dictionary of keyword arguments to pass through to matplotlib's ``title`` fucntion.
        This includes things like fontfamily, fontsize, fontweight, color, etc.

    pylabeladjust_speed: None or float (optional, default=None)
        pylabeladjust speed for adjusting label positioning when doing labels over points. If
        ``label_over_points`` is ``False`` then this will have no effect. If ``None`` then
        a good choice of speed will be approximated from the data.

    pylabeladjust_max_iterations: int (optional, default=500)
        The maximum number of pylabeladjust iterations for adjusting label positioning when
        doing labels over points. If ``label_over_points`` is ``False`` then this will have
        no effect.

    pylabeladjust_adjust_by_size: bool (optional, default=True)
        Whether to adjust the labels based on the size of the rectangles for adjusting label
        positioning when doing labels over points. If ``label_over_points`` is ``False`` then
        this will have no effect.

    pylabeladjust_margin_percentage: float (optional, default=7.5)
        The margin percentage for the repulsion radius for adjusting label positioning when
        doing labels over points. If ``label_over_points`` is ``False`` then this will have no effect.

    pylabeladjust_radius_scale: float (optional, default=1.05)
        The scale factor for the repulsion radius for adjusting label
        positioning when doing labels over points. If ``label_over_points`` is ``False`` then
        this will have no effect.

    label_font_stroke_width: float (optional, default=3)
        The width of the stroke to use when rendering the font. This is used to create an outline
        that distinguishes the text from the background. Larger values will make text more visible 
        against the background at some loss of font legibility. You may need to change this value
        when rendering at particularly high resolutions.

    label_font_outline_alpha: float (optional, default=0.5)
        The alpha value to use when rendering the font outline. This is used to create an outline
        that distinguishes the text from the background. Larger values will make text more visible
        against the background at some loss of font legibility.

    verbose: bool (optional, default=False)
        Print progress as the plot is being created.

    ax: None or matplotlib.axes (optional, default=None)
        If not None, render the plot to this axis, otherwise create a new figure and axis.

    Returns
    -------
    fig: matplotlib.Figure
        The figure that the resulting plot is rendered to.

    ax: matpolotlib.Axes
        The axes contained within the figure that the plot is rendered to.

    """
    # Create the figure
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize, dpi=dpi, constrained_layout=True)
    else:
        fig = ax.get_figure()


    if _can_reach_google_fonts(timeout=5.0):
        if verbose:
            print("Getting any required fonts...")
        # Get any google fonts if required
        get_google_font(font_family)
        get_google_font(font_family.split()[0])
        if title_keywords is not None and "fontfamily" in title_keywords:
            get_google_font(title_keywords["fontfamily"])
            get_google_font(title_keywords["fontfamily"].split()[0])
        if sub_title_keywords is not None and "fontfamily" in sub_title_keywords:
            get_google_font(sub_title_keywords["fontfamily"])
            get_google_font(sub_title_keywords["fontfamily"].split()[0])
    else:
        warn(
            "Cannot reach out Google APIs to download the font you selected. Will fallback on fonts already installed.",
            GoogleAPIUnreachable
        )

    # Apply matplotlib or datashader based on heuristics
    if data_map_coords.shape[0] < 100_000 or force_matplotlib:
        if marker_size_array is not None:
            point_size = marker_size_array * point_size
        ax.scatter(
            *data_map_coords.T,
            c=color_list,
            marker=marker_type,
            s=point_size,
            alpha=alpha,
            edgecolors="none",
        )
    else:
        if marker_size_array is not None or marker_type != "o":
            warn(
                "Adjusting marker type or size cannot be done with datashader; use force_matplotlib=True"
            )
        datashader_scatterplot(
            data_map_coords, color_list, point_size=point_size, ax=ax
        )

    # Create background glow
    if verbose:
        print("Adding glow to scatterplot...")
    if add_glow:
        add_glow_to_scatterplot(
            data_map_coords, color_list, ax, noise_color=noise_color, **glow_keywords
        )

    # Add a mark in the bottom right if provided
    if logo is not None:
        mark_height = (
            (figsize[0] / figsize[1]) * (logo.shape[0] / logo.shape[1]) * logo_width
        )
        ax.imshow(
            logo,
            extent=(0.98 - logo_width, 0.98, 0.02, 0.02 + mark_height),
            transform=ax.transAxes,
        )

    # Find initial placements for text, fix any line crossings, then optimize placements
    if verbose:
        print("Placing labels...")
    ax.autoscale_view()
    if label_locations.shape[0] > 0:

        # Ensure we can look up labels for highlighting
        if highlight_labels is not None:
            highlight = set(highlight_labels)
        else:
            highlight = set([])

        if label_over_points:
            font_scale_factor = np.sqrt(figsize[0] * figsize[1])
            if verbose:
                print("Estimating font size...")
            if label_font_size is None:
                if dynamic_label_size:
                    font_sizes, font_weights = estimate_dynamic_font_size(
                        label_locations,
                        label_text,
                        fontfamily=font_family,
                        linespacing=label_linespacing,
                        expand=(1.0, 1.0),
                        overlap_percentage_allowed=0.66,
                        dynamic_size_array=label_cluster_sizes
                        ** dynamic_label_size_scaling_factor,
                        min_font_size=min_font_size,
                        max_font_size=max_font_size,
                        min_font_weight=min_font_weight,
                        max_font_weight=max_font_weight,
                        ax=ax,
                    )
                    font_size = None
                else:
                    font_size = estimate_font_size(
                        label_locations,
                        label_text,
                        font_scale_factor,
                        fontfamily=font_family,
                        fontweight=font_weight,
                        linespacing=label_linespacing,
                        expand=(1.0, 1.0),
                        overlap_percentage_allowed=0.66,
                        min_font_size=min_font_size,
                        max_font_size=max_font_size,
                        ax=ax,
                    )
                    font_sizes = None
                    font_weights = None
            else:
                font_size = label_font_size
                font_sizes = None
                font_weights = None

            label_text_locations = pylabeladjust_text_locations(
                label_locations,
                label_text,
                fontfamily=font_family,
                font_size=font_size,
                font_sizes=font_sizes,
                font_weights=font_weights,
                linespacing=label_linespacing,
                highlight=highlight,
                highlight_label_keywords=highlight_label_keywords,
                ax=ax,
                fig=fig,
                speed=pylabeladjust_speed,
                max_iterations=pylabeladjust_max_iterations,
                adjust_by_size=pylabeladjust_adjust_by_size,
                margin_percentage=pylabeladjust_margin_percentage,
                radius_scale=pylabeladjust_radius_scale,
            )
        else:
            if verbose:
                print("Creating initial label placements...")
            label_text_locations = initial_text_location_placement(
                label_locations,
                base_radius=label_base_radius,
                theta_stretch=label_direction_bias,
            )
            fix_crossings(label_text_locations, label_locations)

            if verbose:
                print("Estimating font size...")
            font_scale_factor = np.sqrt(figsize[0] * figsize[1])
            if label_font_size is None:
                if dynamic_label_size:
                    font_sizes, font_weights = estimate_dynamic_font_size(
                        label_locations,
                        label_text,
                        fontfamily=font_family,
                        linespacing=label_linespacing,
                        expand=(label_margin_factor, label_margin_factor),
                        overlap_percentage_allowed=0.5,
                        dynamic_size_array=label_cluster_sizes
                        ** dynamic_label_size_scaling_factor,
                        min_font_size=min_font_size,
                        max_font_size=max_font_size,
                        min_font_weight=min_font_weight,
                        max_font_weight=max_font_weight,
                        ax=ax,
                    )
                    font_size = None
                else:
                    font_size = estimate_font_size(
                        label_text_locations,
                        label_text,
                        0.9 * font_scale_factor,
                        fontfamily=font_family,
                        fontweight=font_weight,
                        linespacing=label_linespacing,
                        min_font_size=min_font_size,
                        max_font_size=max_font_size,
                        ax=ax,
                    )
                    font_sizes = None
                    font_weights = None
            else:
                font_size = label_font_size
                font_sizes = None
                font_weights = None

            if verbose:
                print("Adjusting label placements...")
            label_text_locations = adjust_text_locations(
                label_text_locations,
                label_locations,
                label_text,
                fontfamily=font_family,
                font_size=font_size,
                fontweight=font_weight,
                linespacing=label_linespacing,
                highlight=highlight,
                highlight_label_keywords=highlight_label_keywords,
                ax=ax,
                expand=(label_margin_factor, label_margin_factor),
                font_sizes=font_sizes,
                font_weights=font_weights,
            )

        # Build highlight boxes
        if (
            "bbox" in highlight_label_keywords
            and highlight_label_keywords["bbox"] is not None
        ):
            base_bbox_keywords = highlight_label_keywords["bbox"]
        else:
            base_bbox_keywords = None

        if verbose:
            print("Adding labels to the plot...")
        # Add the annotations to the plot
        texts = []
        for i in range(label_locations.shape[0]):
            if base_bbox_keywords is not None:
                bbox_keywords = dict(base_bbox_keywords.items())
                if "fc" not in base_bbox_keywords:
                    if highlight_colors is not None:
                        bbox_keywords["fc"] = highlight_colors[i][:7] + "33"
                    else:
                        bbox_keywords["fc"] = "#cccccc33" if darkmode else "#33333333"
                if "ec" not in base_bbox_keywords:
                    bbox_keywords["ec"] = "none"
            else:
                bbox_keywords = None

            if type(label_text_colors) == str:
                text_color = label_text_colors
            elif label_text_colors:
                text_color = label_text_colors[i]
            elif darkmode:
                text_color = "white"
            else:
                text_color = "black"

            outline_alpha = hex(int(255 * label_font_outline_alpha)).removeprefix('0x')
            outline_color = f"#000000{outline_alpha}" if darkmode else f"#ffffff{outline_alpha}"

            if type(label_arrow_colors) == str:
                arrow_color = label_arrow_colors
            elif label_arrow_colors:
                arrow_color = label_arrow_colors[i]
            elif darkmode:
                arrow_color = "#dddddd"
            else:
                arrow_color = "#333333"

            texts.append(
                ax.annotate(
                    label_text[i],
                    label_locations[i],
                    xytext=label_text_locations[i],
                    ha="center",
                    ma="center",
                    va="center",
                    linespacing=label_linespacing,
                    fontfamily=font_family,
                    arrowprops=(
                        {
                            "arrowstyle": "-",
                            "linewidth": 0.5,
                            "color": arrow_color,
                            **arrowprops,
                        }
                        if not label_over_points
                        else None
                    ),
                    fontsize=(
                        highlight_label_keywords.get("fontsize", font_size)
                        if label_text[i] in highlight
                        else font_size
                    )
                    if font_sizes is None
                    else font_sizes[i],
                    path_effects=(
                        [
                            patheffects.Stroke(linewidth=label_font_stroke_width, foreground=outline_color),
                            patheffects.Normal(),
                        ]
                        if label_over_points
                        else None
                    ),
                    bbox=bbox_keywords if label_text[i] in highlight else None,
                    color=text_color,
                    fontweight=(
                        highlight_label_keywords.get("fontweight", font_weight)
                        if label_text[i] in highlight
                        else (font_weights[i] if font_weights is not None else font_weight)
                    ),
                )
            )

        # Ensure we have plot bounds that meet the newly place annotations
        coords = get_2d_coordinates(texts)
        x_min, y_min = ax.transData.inverted().transform(
            (coords[:, [0, 2]].copy().min(axis=0))
        )
        x_max, y_max = ax.transData.inverted().transform(
            (coords[:, [1, 3]].copy().max(axis=0))
        )
        width = x_max - x_min
        height = y_max - y_min
        x_min -= 0.05 * width
        x_max += 0.05 * width
        y_min -= 0.05 * height
        y_max += 0.05 * height
    else:
        x_min, y_min = data_map_coords.min(axis=0)
        x_max, y_max = data_map_coords.max(axis=0)
        width = x_max - x_min
        height = y_max - y_min
        x_min -= 0.05 * width
        x_max += 0.05 * width
        y_min -= 0.05 * height
        y_max += 0.05 * height

    # decorate the plot
    if verbose:
        print("Decorating plot...")
    ax.set(xticks=[], yticks=[])
    if sub_title is not None:
        if sub_title_keywords is not None:
            keyword_args = {
                "fontweight": "light",
                "color": "gray",
                "fontsize": (1.6 * font_scale_factor),
                "fontfamily": font_family,
                "fontweight": font_weight,
                **sub_title_keywords,
            }
        else:
            keyword_args = {
                "fontweight": "light",
                "color": "gray",
                "fontsize": (1.6 * font_scale_factor),
                "fontfamily": font_family,
                "fontweight": font_weight,
            }
        axis_title = ax.set_title(
            sub_title,
            loc="left",
            va="baseline",
            fontdict=keyword_args,
        )
        sup_title_y_value = (
            ax.transAxes.inverted().transform(
                get_2d_coordinates([axis_title])[0, [0, 3]]
            )[1]
            + 1e-4
        )
    else:
        sup_title_y_value = 1.00

    if title is not None:
        if title_keywords is not None:
            keyword_args = {
                "color": "white" if darkmode else "black",
                "ha": "left",
                "va": "bottom",
                "fontweight": 900,
                "fontsize": int(3.2 * font_scale_factor),
                "fontfamily": font_family,
                **title_keywords,
            }
        else:
            keyword_args = {
                "color": "white" if darkmode else "black",
                "ha": "left",
                "va": "bottom",
                "fontweight": 900,
                "fontsize": int(3.2 * font_scale_factor),
                "fontfamily": font_family,
            }
        fig.suptitle(
            title, x=0.0, y=sup_title_y_value, transform=ax.transAxes, **keyword_args
        )

    ax_x_min, ax_x_max = ax.get_xlim()
    ax_y_min, ax_y_max = ax.get_ylim()
    ax.set_xlim(min(x_min, ax_x_min), max(x_max, ax_x_max))
    ax.set_ylim(min(y_min, ax_y_min), max(y_max, ax_y_max))
    for spine in ax.spines.values():
        spine.set_edgecolor("#555555" if darkmode else "#aaaaaa")
    ax.set_aspect("auto")

    if darkmode:
        ax.set(facecolor="black")
        fig.set(facecolor="black")

    return fig, ax



import numpy as np
from sklearn.metrics import pairwise_distances
from sklearn.preprocessing import MinMaxScaler
from pylabeladjust import adjust_texts

from datamapplot.overlap_computations import (
    get_2d_coordinates,
    overlap_intervals,
    text_line_overlaps,
    intersect,
)

from matplotlib import pyplot as plt


def row_norm(an_array):
    result = np.empty(an_array.shape[0], dtype=np.float64)
    for i in range(an_array.shape[0]):
        result[i] = np.sqrt(np.sum(an_array[i] ** 2))
    return result


def fix_crossings(text_locations, label_locations, n_iter=3):
    # Find crossing lines and swap labels; repeat as required
    for n in range(n_iter):
        for i in range(text_locations.shape[0]):
            for j in range(text_locations.shape[0]):
                if intersect(
                    text_locations[i],
                    label_locations[i],
                    text_locations[j],
                    label_locations[j],
                ):
                    swap = text_locations[i].copy()
                    text_locations[i] = text_locations[j]
                    text_locations[j] = swap


def initial_text_location_placement(
    label_locations, base_radius=None, theta_stretch=None
):
    # Heuristic to choose how many rings of text labels to use
    n_label_rings = (label_locations.shape[0] // 32) + 1

    # Find a center for label locations, ring radii, and how much to stretch theta; all heuristics
    mean_label_location = np.sum(label_locations, axis=0) / label_locations.shape[0]
    recentered_label_locations = label_locations - mean_label_location
    if base_radius is None:
        base_radius = np.max(row_norm(recentered_label_locations)) + 0.25 * np.mean(
            row_norm(recentered_label_locations)
        )
    if theta_stretch is None:
        theta_stretch = np.clip(((label_locations.shape[0] - 30) / 10), 1.0, 1.66)
    recentered_label_locations = (
        recentered_label_locations / row_norm(recentered_label_locations)[:, None]
    )

    # Spread thetas out such that we have smaller angles near the east-west points, larger at north sound
    # (since text is horizontal)
    label_thetas = np.arctan2(
        recentered_label_locations.T[0], recentered_label_locations.T[1]
    )
    xs = (
        np.linspace(0, 1, len(label_thetas) // 4 + (n_label_rings + 2)) ** theta_stretch
    )
    positive_thetas = np.hstack([xs * np.pi / 2, (np.pi - xs[::-1] * np.pi / 2)[1:]])
    uniform_thetas = np.hstack([positive_thetas, -positive_thetas[1:-1]])

    # Choose which ring to be in and set radii
    ring_choice = np.concatenate(
        [np.arange(n_label_rings), np.arange(n_label_rings - 2, 0, -1)]
    )
    ring_choice = np.tile(
        ring_choice, (uniform_thetas.shape[0] // ring_choice.shape[0]) + 1
    )
    base_radii = np.asarray(
        [
            base_radius * (1.33 ** (ring_choice[i]))
            for i in range(uniform_thetas.shape[0])
        ]
    )

    # Find an optimal rotation to match the existing label locations
    optimal_rotation = 0.0
    min_score = np.inf
    for rotation in np.linspace(
        -np.pi / int(len(label_thetas) + 5), np.pi / int(len(label_thetas) + 5), 32
    ):
        test_label_locations = np.vstack(
            [
                base_radii * np.cos(uniform_thetas + rotation),
                base_radii * np.sin(uniform_thetas + rotation),
            ]
        ).T
        score = np.sum(
            pairwise_distances(
                recentered_label_locations, test_label_locations, metric="cosine"
            ).min(axis=1)
        )
        if score < min_score:
            min_score = score
            optimal_rotation = rotation

    uniform_thetas += optimal_rotation
    uniform_label_locations = np.vstack(
        [base_radii * np.cos(uniform_thetas), base_radii * np.sin(uniform_thetas)]
    ).T

    # Sort labels by radius of the label location and pick the closest position in order;
    # This works surprisingly well
    order = np.argsort(-row_norm(label_locations - mean_label_location))
    taken = set([])
    adjustment_dict_alt = {}
    for i in order:
        candidates = list(set(range(uniform_label_locations.shape[0])) - taken)
        candidate_distances = pairwise_distances(
            [recentered_label_locations[i]],
            uniform_label_locations[candidates],
            metric="cosine",
        )
        selection = candidates[np.argmin(candidate_distances[0])]
        adjustment_dict_alt[i] = selection
        taken.add(selection)

    result = (
        np.asarray(
            [
                uniform_label_locations[adjustment_dict_alt[i]]
                for i in sorted(adjustment_dict_alt.keys())
            ]
        )
        + mean_label_location
    )

    return result


def estimate_font_size(
    text_locations,
    label_text,
    initial_font_size,
    fontfamily="Roboto",
    fontweight=400,
    linespacing=0.95,
    expand=(1.5, 1.5),
    overlap_percentage_allowed=0.5,
    min_font_size=3.0,
    max_font_size=16.0,
    ax=None,
):
    if ax is None:
        ax = plt.gca()

    font_size = initial_font_size
    overlap_percentage = 1.0
    while overlap_percentage > overlap_percentage_allowed and font_size > min_font_size:
        texts = [
            ax.text(
                *text_locations[i],
                label_text[i],
                ha="center",
                ma="center",
                va="center",
                linespacing=linespacing,
                alpha=0.0,
                fontfamily=fontfamily,
                fontweight=fontweight,
                fontsize=font_size,
            )
            for i in range(text_locations.shape[0])
        ]
        coords = get_2d_coordinates(texts, expand=expand)
        xoverlaps = overlap_intervals(
            coords[:, 0], coords[:, 1], coords[:, 0], coords[:, 1]
        )
        xoverlaps = xoverlaps[xoverlaps[:, 0] != xoverlaps[:, 1]]
        yoverlaps = overlap_intervals(
            coords[:, 2], coords[:, 3], coords[:, 2], coords[:, 3]
        )
        yoverlaps = yoverlaps[yoverlaps[:, 0] != yoverlaps[:, 1]]
        overlaps = yoverlaps[(yoverlaps[:, None] == xoverlaps).all(-1).any(-1)]
        overlap_percentage = len(overlaps) / (2 * text_locations.shape[0])
        # remove texts
        for t in texts:
            t.remove()

        font_size = 0.9 * font_size

    return font_size


def estimate_dynamic_font_size(
    text_locations,
    label_text,
    fontfamily="DejaVu Sans",
    linespacing=0.95,
    expand=(1.5, 1.5),
    overlap_percentage_allowed=0.5,
    dynamic_size_array=None,
    min_font_size=4.0,
    max_font_size=24.0,
    min_font_weight=200,
    max_font_weight=500,
    ax=None,
):
    if ax is None:
        ax = plt.gca()

    overlap_percentage = 1.0
    current_max_font_size = max_font_size
    weight_scaler = MinMaxScaler(feature_range=(min_font_weight, max_font_weight))
    font_weights = np.squeeze(
        weight_scaler.fit_transform(dynamic_size_array.reshape(-1, 1))
    )
    while (
        overlap_percentage > overlap_percentage_allowed
        and current_max_font_size > min_font_size
    ):
        size_scaler = MinMaxScaler(feature_range=(min_font_size, current_max_font_size))
        font_sizes = np.squeeze(
            size_scaler.fit_transform(dynamic_size_array.reshape(-1, 1))
        )
        texts = [
            ax.text(
                *text_locations[i],
                label_text[i],
                ha="center",
                ma="center",
                va="center",
                linespacing=linespacing,
                alpha=0.0,
                fontfamily=fontfamily,
                fontsize=font_sizes[i],
                fontweight=font_weights[i],
            )
            for i in range(text_locations.shape[0])
        ]
        coords = get_2d_coordinates(texts, expand=expand)
        xoverlaps = overlap_intervals(
            coords[:, 0], coords[:, 1], coords[:, 0], coords[:, 1]
        )
        xoverlaps = xoverlaps[xoverlaps[:, 0] != xoverlaps[:, 1]]
        yoverlaps = overlap_intervals(
            coords[:, 2], coords[:, 3], coords[:, 2], coords[:, 3]
        )
        yoverlaps = yoverlaps[yoverlaps[:, 0] != yoverlaps[:, 1]]
        overlaps = yoverlaps[(yoverlaps[:, None] == xoverlaps).all(-1).any(-1)]
        overlap_percentage = len(overlaps) / (2 * text_locations.shape[0])
        # remove texts
        for t in texts:
            t.remove()

        current_max_font_size = 0.9 * current_max_font_size

    return font_sizes, font_weights


def adjust_text_locations(
    text_locations,
    label_locations,
    label_text,
    font_size=12,
    fontfamily="DejaVu Sans",
    fontweight=400,
    linespacing=0.95,
    expand=(1.5, 1.5),
    max_iter=100,
    highlight=frozenset([]),
    highlight_label_keywords={},
    font_sizes=None,
    font_weights=None,
    ax=None,
):
    if ax is None:
        ax = plt.gca()

    # Add text to the axis and set up for optimization
    new_text_locations = text_locations.copy()
    texts = [
        ax.text(
            *new_text_locations[i],
            label_text[i],
            ha="center",
            ma="center",
            va="center",
            linespacing=linespacing,
            alpha=0.0,
            fontfamily=fontfamily,
            fontsize=(
                highlight_label_keywords.get("fontsize", font_size)
                if label_text[i] in highlight
                else font_size
            )
            if font_sizes is None
            else font_sizes[i],
            fontweight=(
                "bold"
                if label_text[i] in highlight
                else (font_weights[i] if font_weights is not None else fontweight)
            ),
        )
        for i in range(label_locations.shape[0])
    ]
    coords = get_2d_coordinates(texts, expand=expand)
    xoverlaps = overlap_intervals(
        coords[:, 0], coords[:, 1], coords[:, 0], coords[:, 1]
    )
    xoverlaps = xoverlaps[xoverlaps[:, 0] != xoverlaps[:, 1]]
    yoverlaps = overlap_intervals(
        coords[:, 2], coords[:, 3], coords[:, 2], coords[:, 3]
    )
    yoverlaps = yoverlaps[yoverlaps[:, 0] != yoverlaps[:, 1]]
    overlaps = yoverlaps[(yoverlaps[:, None] == xoverlaps).all(-1).any(-1)]
    tight_coords = get_2d_coordinates(texts, expand=(0.9, 0.9))
    bottom_lefts = ax.transData.inverted().transform(tight_coords[:, [0, 2]])
    top_rights = ax.transData.inverted().transform(tight_coords[:, [1, 3]])
    coords_in_dataspace = np.vstack(
        [bottom_lefts.T[0], top_rights.T[0], bottom_lefts.T[1], top_rights.T[1]]
    ).T
    box_line_overlaps = text_line_overlaps(
        text_locations, label_locations, coords_in_dataspace
    )
    n_iter = 0

    # While we have overlaps, tweak the label positions
    while (len(overlaps) > 0 or len(box_line_overlaps) > 0) and n_iter < max_iter:
        # Check for text boxes overlapping each other
        coords = get_2d_coordinates(texts, expand=expand)
        xoverlaps = overlap_intervals(
            coords[:, 0], coords[:, 1], coords[:, 0], coords[:, 1]
        )
        xoverlaps = xoverlaps[xoverlaps[:, 0] != xoverlaps[:, 1]]
        yoverlaps = overlap_intervals(
            coords[:, 2], coords[:, 3], coords[:, 2], coords[:, 3]
        )
        yoverlaps = yoverlaps[yoverlaps[:, 0] != yoverlaps[:, 1]]
        overlaps = yoverlaps[(yoverlaps[:, None] == xoverlaps).all(-1).any(-1)]
        recentered_locations = new_text_locations - label_locations.mean(axis=0)
        radii = np.linalg.norm(recentered_locations, axis=1)
        thetas = np.arctan2(recentered_locations.T[1], recentered_locations.T[0])
        for left, right in overlaps:
            # adjust thetas
            direction = thetas[left] - thetas[right]
            if direction > np.pi or direction < -np.pi:
                thetas[left] -= 0.005 * np.sign(direction)
                thetas[right] += 0.005 * np.sign(direction)
            else:
                thetas[left] += 0.005 * np.sign(direction)
                thetas[right] -= 0.005 * np.sign(direction)

        # Check for indicator lines crossing text boxes
        recentered_locations = np.vstack(
            [radii * np.cos(thetas), radii * np.sin(thetas)]
        ).T
        new_text_locations = recentered_locations + label_locations.mean(axis=0)
        fix_crossings(new_text_locations, label_locations)
        for i, text in enumerate(texts):
            text.set_position(new_text_locations[i])

        tight_coords = get_2d_coordinates(texts, expand=expand)
        bottom_lefts = ax.transData.inverted().transform(tight_coords[:, [0, 2]])
        top_rights = ax.transData.inverted().transform(tight_coords[:, [1, 3]])
        coords_in_dataspace = np.vstack(
            [bottom_lefts.T[0], top_rights.T[0], bottom_lefts.T[1], top_rights.T[1]]
        ).T
        box_line_overlaps = text_line_overlaps(
            new_text_locations, label_locations, coords_in_dataspace
        )
        recentered_locations = new_text_locations - label_locations.mean(axis=0)
        radii = np.linalg.norm(recentered_locations, axis=1)
        thetas = np.arctan2(recentered_locations.T[1], recentered_locations.T[0])

        for i, j in box_line_overlaps:
            direction = np.arctan2(
                np.sum(coords_in_dataspace[i, 2:]) / 2.0 - label_locations[j, 1],
                np.sum(coords_in_dataspace[i, :2]) / 2.0 - label_locations[j, 0],
            ) - np.arctan2(
                text_locations[j, 1] - label_locations[j, 1],
                text_locations[j, 0] - label_locations[j, 0],
            )
            if direction > np.pi or direction < -np.pi:
                thetas[i] -= 0.005 * np.sign(direction)
                thetas[j] += 0.0025 * np.sign(direction)
            else:
                thetas[i] += 0.005 * np.sign(direction)
                thetas[j] -= 0.0025 * np.sign(direction)

        radii *= 1.003

        recentered_locations = np.vstack(
            [radii * np.cos(thetas), radii * np.sin(thetas)]
        ).T
        new_text_locations = recentered_locations + label_locations.mean(axis=0)
        fix_crossings(new_text_locations, label_locations)
        for i, text in enumerate(texts):
            text.set_position(new_text_locations[i])

        n_iter += 1

    return new_text_locations


def pylabeladjust_text_locations(
    label_locations,
    label_text,
    font_size=12,
    font_sizes=None,
    font_weights=None,
    fontfamily="DejaVu Sans",
    linespacing=0.95,
    highlight=frozenset([]),
    highlight_label_keywords={},
    speed=None,
    max_iterations=500,
    adjust_by_size=True,
    margin_percentage=7.5,
    radius_scale=1.05,
    ax=None,
    fig=None,
):
    if ax is None:
        ax = plt.gca()

    if fig is None:
        fig = plt.gcf()

    if speed is None:
        data_radius = np.max(row_norm(label_locations))
        speed = data_radius / 125.0

    # Add text to the axis and set up for optimization
    new_text_locations = label_locations.copy()
    texts = [
        ax.text(
            *new_text_locations[i],
            label_text[i],
            ha="center",
            ma="center",
            va="center",
            linespacing=linespacing,
            alpha=0.0,
            fontfamily=fontfamily,
            fontsize=(
                highlight_label_keywords.get("fontsize", font_size)
                if label_text[i] in highlight
                else font_size
            )
            if font_sizes is None
            else font_sizes[i],
            fontweight=(
                "bold"
                if label_text[i] in highlight
                else (font_weights[i] if font_weights is not None else "normal")
            ),
        )
        for i in range(label_locations.shape[0])
    ]
    fig.canvas.draw()
    rectangles_adjusted = adjust_texts(
        texts,
        speed=speed,
        max_iterations=max_iterations,
        adjust_by_size=adjust_by_size,
        margin=margin_percentage,
        margin_type="percentage",
        radius_scale=radius_scale,
    )
    return rectangles_adjusted[["x_center", "y_center"]].values



import numpy as np
import numba

from scipy.interpolate import splprep, splev


@numba.njit()
def circumradius(points):
    bc = points[1:] - points[0]
    d = 2 * (bc[0, 0] * bc[1, 1] - bc[0, 1] * bc[1, 0])
    if d == 0:
        return 0
    b_norm = bc[0, 0] * bc[0, 0] + bc[0, 1] * bc[0, 1]
    c_norm = bc[1, 0] * bc[1, 0] + bc[1, 1] * bc[1, 1]
    ux = (bc[1, 1] * b_norm - bc[0, 1] * c_norm) / d
    uy = (bc[0, 0] * c_norm - bc[1, 0] * b_norm) / d
    return np.sqrt(ux * ux + uy * uy)


@numba.njit(locals={"candidate_idx": numba.uint64})
def find_boundary_candidates(points, simplices, alpha=0.1):
    candidates = np.full((simplices.shape[0] * 3, 2), -1, dtype=np.int32)
    candidate_idx = 0
    for simplex in simplices:
        if circumradius(points[simplex]) < alpha:
            candidates[candidate_idx] = (simplex[0], simplex[1])
            candidates[candidate_idx + 1] = (simplex[0], simplex[2])
            candidates[candidate_idx + 2] = (simplex[1], simplex[2])
            candidate_idx += 3
    return candidates[:candidate_idx]

@numba.njit()
def boundary_from_candidates(boundary_candidates):
    occurrence_counts = {(np.int32(0), np.int32(0)):0 for i in range(0)}
    for candidate in boundary_candidates:
        tuple_candidate = (candidate[0], candidate[1])
        if tuple_candidate in occurrence_counts:
            occurrence_counts[tuple_candidate] += 1
        else:
            occurrence_counts[tuple_candidate] = 1

    return set([x for x in occurrence_counts if occurrence_counts[x] == 1])

@numba.njit()
def build_polygons(boundary):
    polygons = []
    search_set = boundary.copy()
    sequence = list(search_set.pop())
    while len(search_set) > 0:
        to_find = sequence[-1]
        for link in search_set:
            if link[0] == to_find:
                sequence.append(link[1])
                search_set.remove(link)
                break
            elif link[1] == to_find:
                sequence.append(link[0])
                search_set.remove(link)
                break
        else:
            polygons.append(sequence.copy())
            sequence = list(search_set.pop())

    polygons.append(sequence)
    return polygons

def create_boundary_polygons(points, simplices, alpha=0.1):
    simplices.sort(axis=1)
    boundary_candidates = find_boundary_candidates(points, simplices, alpha=alpha)
    boundary = boundary_from_candidates(boundary_candidates)
    if len(boundary) == 0:
        raise ValueError("The value of polygon_alpha was too low, and no boundary was formed. Try increasing polygon_alpha.")
    polygons = build_polygons(boundary)

    result = [
        np.empty((len(sequence) + 1, 2), dtype=np.float32) for sequence in polygons
    ]
    for s, sequence in enumerate(polygons):
        for i, n in enumerate(sequence):
            result[s][i] = points[n]
        result[s][-1] = points[sequence[0]]

    return result


def smooth_polygon(p, point_multipler=4, spline_coeff=0.0001):
    dist = np.sqrt(np.sum((p[:-2] - p[1:-1]) ** 2, axis=1))
    dist_along = np.concatenate(([0], dist.cumsum()))
    spline, u = splprep(p[:-1].T, u=dist_along, s=spline_coeff, per=True)

    interp_d = np.linspace(dist_along[0], dist_along[-1], len(p) * point_multipler)
    interp_x, interp_y = splev(interp_d, spline)

    return np.vstack([interp_x, interp_y]).T



import pandas as pd
import numpy as np
from typing import Tuple, Optional

_NAME_TO_FREQ_CODE = {
    "year": "YS",
    "y": "YS",
    "quarter": "QS",
    "Q": "QS",
    "q": "QS",
    "month": "MS",
    "M": "MS",
    "day": "D",
    "d": "D",
    "hour": "H",
    "h": "H",
    "minute": "T",
    "min": "T",
    "m": "T",
    "second": "S",
    "sec": "S",
    "s": "S",
}

_NAME_TO_PERIOD_CODE = {
    "year": "Y",
    "y": "Y",
    "quarter": "Q",
    "Q": "Q",
    "q": "Q",
    "month": "M",
    "M": "",
    "day": "D",
    "d": "D",
    "hour": "h",
    "h": "h",
    "minute": "min",
    "min": "min",
    "m": "min",
    "second": "s",
    "sec": "s",
    "s": "s",
}


def generate_bins_from_numeric_data(
    data: pd.Series,
    n_bins: int = 20,
    histogram_range: Optional[Tuple[float, float]] = None,
):
    """
    Generate bins for numeric data.

    Parameters
    ----------
    data : pd.Series or np.ndarray
        The data to bin.
    n_bins : int
        The number of bins to generate.

    Returns
    -------
    pd.DataFrame
    A DataFrame with the following columns:
        - id: The bin id.
        - count: The number of data points in the bin.
        - indices: The indices of the data points in the bin.
        - min_value: The minimum value of the bin.
        - max_value: The maximum value of the bin.
        - mean_value: The mean value of the bin.

    pd.Series
    A Series with the bin ids, one per entry in the data.
    """
    if isinstance(data, np.ndarray):
        data = pd.Series(data)

    if histogram_range is None:
        bins_ids, bin_edges = pd.cut(data, bins=n_bins, labels=False, retbins=True)
    else:
        from_value, to_value = histogram_range
        if isinstance(from_value, pd.Timestamp):
            bins = pd.date_range(from_value, to_value, periods=n_bins)
        else:
            bins = np.linspace(from_value, to_value, n_bins)
        bins_ids, bin_edges = pd.cut(data, bins, labels=False, retbins=True)
        if not np.all(np.isfinite(bins_ids)):
            bins_ids = np.where(data < from_value, 0, bins_ids)
            bins_ids = np.where(data > to_value, len(bin_edges) - 1, bins_ids)

    all_bins = pd.DataFrame({"id": range(n_bins)})

    bin_data = (
        data.groupby(bins_ids, group_keys=True)
        .agg(count="count", indices=lambda x: list(x.index))
        .reset_index(names=["id"])
    )

    bin_data = all_bins.merge(bin_data, on="id", how="left")

    bin_data["min_value"] = bin_edges[:-1]
    bin_data["max_value"] = bin_edges[1:]
    # lower bin plus difference over two to allow to dates/times to work with averages as well
    bin_data["mean_value"] = bin_edges[:-1] + (bin_edges[1:] - bin_edges[:-1]) / 2
    bin_data["count"] = bin_data["count"].fillna(0)
    bin_data["indices"] = bin_data["indices"].apply(
        lambda x: x if isinstance(x, list) else []
    )

    return bin_data, bins_ids.astype(np.int16).rename("bin_id")


def generate_bins_from_categorical_data(
    data: pd.Series,
    max_bins: int = 20,
    histogram_range: Optional[Tuple[float, float]] = None,
):
    """
    Generate bins for categorical data.

    Parameters
    ----------
    data : pd.Series or np.ndarray
        The data to bin.
    max_bins : int
        The maximum number of bins to generate.

    Returns
    -------
    pd.DataFrame
    A DataFrame with the following columns:
        - id: The bin id.
        - count: The number of data points in the bin.
        - indices: The indices of the data points in the bin.
        - min_value: The minimum value of the bin.
        - max_value: The maximum value of the bin.
        - mean_value: The mean value of the bin.

    pd.Series
    A Series with the bin ids, one per entry in the data.
    """
    if isinstance(data, np.ndarray):
        data = pd.Series(data)
    top_values = data.value_counts().head(max_bins - 1).index
    bins_ids = data.apply(lambda x: x if x in top_values else "Other")
    bin_data = (
        data.groupby(bins_ids)
        .agg(count="count", indices=lambda x: list(x.index))
        .reset_index(names=["label"])
        .reset_index(names=["id"])
    )

    bin_data["min_value"] = bin_data["id"]
    bin_data["max_value"] = bin_data["id"]
    bin_data["mean_value"] = bin_data["label"]
    bin_data["count"] = bin_data["count"].fillna(0)
    bin_data["indices"] = bin_data["indices"].apply(
        lambda x: x if isinstance(x, list) else []
    )

    return bin_data, bins_ids.map(dict(bin_data[["label", "id"]].values)).astype(
        np.int16
    ).rename("bin_id")


def generate_bins_from_temporal_data(
    data: pd.Series,
    group_by: str = "year",
    histogram_range: Optional[Tuple[float, float]] = None,
):
    """
    Generate bins for temporal data.

    Parameters
    ----------
    data : pd.Series or np.ndarray
        The data to bin.
    group_by : str
        The temporal unit to group by. Can be one of "year", "quarter", "month", "day", "hour", "minute", or "second".

    Returns
    -------
    pd.DataFrame
    A DataFrame with the following columns:
        - id: The bin id.
        - count: The number of data points in the bin.
        - indices: The indices of the data points in the bin.
        - min_value: The minimum value of the bin.
        - max_value: The maximum value of the bin.
        - mean_value: The mean value of the bin.

    pd.Series
    A Series with the bin ids, one per entry in the data.
    """
    if isinstance(data, np.ndarray):
        data = pd.Series(data)
    frequency_code = _NAME_TO_FREQ_CODE.get(group_by, group_by)
    period_code = _NAME_TO_PERIOD_CODE[group_by]

    if histogram_range is None:
        from_date = data.min().to_period(period_code).to_timestamp()
        to_date = (data.max().to_period(period_code) + 1).to_timestamp()
    else:
        from_date, to_date = histogram_range

    bins_ids, bin_edges = pd.cut(
        data,
        pd.date_range(from_date, to_date, freq=frequency_code),
        labels=False,
        retbins=True,
    )
    if not np.all(np.isfinite(bins_ids)):
        bins_ids = np.where(data <= from_date, 0, bins_ids)
        bins_ids = np.where(data >= to_date, len(bin_edges) - 1, bins_ids)
        bins_ids = pd.Series(bins_ids)

    all_bins = pd.DataFrame({"id": range(len(bin_edges) - 1)})

    bin_data = (
        data.groupby(bins_ids, group_keys=True)
        .agg(count="count", indices=lambda x: list(x.index))
        .reset_index(names=["id"])
    )

    bin_data = all_bins.merge(bin_data, on="id", how="left")

    bin_data["min_value"] = bin_edges[:-1]
    bin_data["max_value"] = bin_edges[1:]
    # lower bin plus difference over two to allow to dates/times to work with averages as well
    bin_data["mean_value"] = bin_edges[:-1] + (bin_edges[1:] - bin_edges[:-1]) / 2
    bin_data["count"] = bin_data["count"].fillna(0)
    bin_data["indices"] = bin_data["indices"].apply(
        lambda x: x if isinstance(x, list) else []
    )

    return bin_data, bins_ids.fillna(0).astype(np.int16).rename("bin_id")
